"""This Python code is an automatically generated wrapper
for Fortran code made by 'fmodpy'. The original documentation
for the Fortran source code follows.
"""

import os
import ctypes
import numpy

# --------------------------------------------------------------------
#               CONFIGURATION
#
_verbose = True
_fort_compiler = "gfortran"
_shared_object_name = "vtmop_libe.so"
_this_directory = os.path.dirname(os.path.abspath(__file__))
_path_to_lib = os.path.join(_this_directory, _shared_object_name)
_compile_options = ["-fPIC", "-shared", "-O3", "-std=legacy"]
_ordered_dependencies = [
    "blas.f",
    "shared_modules.f90",
    "qnstop.f90",
    "lapack.f",
    "linear_shepard.f90",
    "slatec.f",
    "delsparse.f90",
    "vtmop.f90",
    "vtmop_libe.f90",
    "vtmop_libe_c_wrapper.f90", ]
#
# --------------------------------------------------------------------
#               AUTO-COMPILING
#
# Try to import the existing object. If that fails, recompile and then try.
try:
    clib = ctypes.CDLL(_path_to_lib)
except Exception:
    # Remove the shared object if it exists, because it is faulty.
    if os.path.exists(_shared_object_name):
        os.remove(_shared_object_name)
    # Compile a new shared object.
    _command = " ".join([_fort_compiler] + _compile_options + ["-o", _shared_object_name] + _ordered_dependencies)
    if _verbose:
        print("Running system command with arguments")
        print("  ", _command)
    # Run the compilation command.
    import subprocess

    subprocess.run(_command, shell=True, cwd=_this_directory)
    # Import the shared object file as a C library with ctypes.
    clib = ctypes.CDLL(_path_to_lib)
# --------------------------------------------------------------------


class vtmop_libe_mod:
    """"""

    # ----------------------------------------------
    # Wrapper for the Fortran subroutine VTMOP_LIBE_INIT

    def vtmop_libe_init(self, d, p, lb, ub, lopt_budget, decay, des_tol, eps, epsw, obj_tol, min_radf, trust_radf,
                        obj_bounds, pmode):
        """! This subroutine initializes a VTMOP object for tracking the adaptive
        ! weighting scheme described in
        !
        ! Deshpande, Shubhangi, Layne T. Watson, and Robert A. Canfield.
        ! "Multiobjective optimization using an adaptive weighting scheme."
        ! Optimization Methods and Software 31.1 (2016): 110-133.
        !
        ! This is a wrapper for the subroutine VTMOP_INIT in VTMOP_MOD.
        !
        !
        ! On input:
        !
        ! D is the dimension of the design space.
        !
        ! P is the dimension of the objective space.
        !
        ! LB(1:D) is the real vector of lower bound constraints for the
        !    D design variables.
        !
        ! UB(1:D) is the real vector of upper bound constraints for the
        !    D design variables.
        !
        !
        ! On output:
        !
        ! IERR is an integer error flag.
        !
        ! Hundreds digit:
        !  000 : Normal output. Successful initialization of VTMOP object.
        !
        !  1xx : Errors detected.
        !   Tens digit:
        !     11x : The input parameters contained illegal dimensions or values.
        !       Ones digit:
        !         110 : D (design dimension) must be a positive integer.
        !         111 : P (objective dimension) must be at least two.
        !         112 : The lead dimension of LB(:) must match D.
        !         113 : The lead dimension of UB(:) must match D.
        !         114 : LB(:) must be elementwise strictly less than UB(:) - DES_TOL.
        !     12x : The optional dummy arguments contained illegal values.
        !       Ones digit:
        !         123 : LOPT_BUDGET must be positive.
        !         124 : DECAY must be in the range (EPS, 1-EPS).
        !         125 : TRUST_RADF must be larger than or equal to MIN_RADF.
        !         129 : If OBJ_BOUNDS is given, then it must have dimensions P by 2
        !               and all OBJ_BOUNDS(:,1) < OBJ_BOUNDS(:,2).
        !     13x : A memory allocation error has occurred.
        !       Ones digit:
        !         130 : A memory allocation error occurred.
        !         131 : A memory deallocation error occurred.
        !
        !  9xx : A checkpointing error has occurred.
        !    901 : WARNING: the VTMOP object was successfully recovered from the
        !          checkpoint but does not match the input data.
        !
        !  The following error codes are returned by VTMOP_CHKPT_NEW. Further details
        !  can be found in the header for VTMOP_CHKPT_NEW.
        !    91x : The VTMOP passed to the checkpoint was invalid.
        !    92x : Error creating the checkpoint file.
        !
        !  The following error codes are returned by VTMOP_CHKPT_RECOVER, when
        !  VTMOP_INIT is called in recovery mode. Further details can be found in
        !  the header for VTMOP_CHKPT_RECOVER.
        !    95x : Error reading data from the checkpoint file.
        !    96x : A memory management error occurred during recovery.
        !
        !
        ! Optional input arguments.
        !
        ! LOPT_BUDGET is an integer input, which specifies the budget for the
        !    local optimization subroutine. The default value for LOPT_BUDGET is
        !    2500 surrogate evaluations. Note that this value is not saved during
        !    checkpointing, and must be reset by the user when recovery mode is
        !    active, whenever a non-default value is desired.
        !
        ! DECAY is a real input specifying the decay rate for the local
        !    trust region (LTR) radius. This value affects how many times an
        !    isolated point can be the center of a LTR before it is discarded.
        !    By default, DECAY = 0.5.
        !
        ! DES_TOL is the tolerance for the design space. A design point that
        !    is within DES_TOL of an evaluated design point will not be reevaluated.
        !    The default value for DES_TOL is the square-root of the working precision
        !    EPS. Note that any value that is smaller than the working precsion EPS
        !    will be ignored and EPS will be used.
        !
        ! EPS is a real input, which specifies the working precision of the
        !    machine. The default value for EPS is SQRT(EPSILON), where EPSILON
        !    is the unit roundoff. Note that if the value supplied is smaller than
        !    the default value then the default value will be used.
        !
        ! EPSW is a small positive number, which is used as the fudge factor for
        !    zero-valued weights. A zero-valued weight does not guarantee Pareto
        !    optimality. Therefore, all zero weights are set to EPSW. The appropriate
        !    value of EPSW is problem dependent. By default, EPSW is the fourth root
        !    of EPSILON (the unit roundoff). Note that any value that is smaller
        !    than SQRT(EPSILON) is ignored and SQRT(EPSILON) will be used.
        !
        ! OBJ_TOL is the tolerance for the objective space. An objective point
        !    that is within OBJ_TOL of being dominated by another objective point
        !    will be treated as such. The default value of OBJ_TOL is the
        !    square-root of EPS. This value should be strictly greater than the
        !    value of EPS. Note that any value that is smaller than the
        !    working precsion EPS will be ignored and EPS will be used.
        !
        ! OBJ_BOUNDS(1:P,1:2) is an optional real P by 2 array, whose first column
        !    is a list of lower bounds and whose second column is a list of upper
        !    bounds on the range of interesting objective values. When present,
        !    this value is used to prune the list of potential LTR centers, so
        !    that only objective values in the specified range are used when
        !    looking for gaps in the Pareto. In particular, an objective value
        !    F(x) will be considered for a LTR center if and only if
        !    OBJ_BOUNDS(I,1) .LE. F_I(x) .LE. OBJ_BOUNDS(I,2) for all I = 1, ..., P.
        !    By default, there are no bounds on the interesting range. Note that
        !    this value is intentionally not saved during checkpointing, and must
        !    be reset by the user when recovery mode is active, whenever a
        !    non-default value is desired. This is the only input, for which
        !    changing the value after loading from a previous checkpoint is not
        !    ill-advised.
        !
        ! MIN_RADF is the smallest value for the fraction r defining the trust region
        !    box dimensions r * (UB - LB), before an isolated point is abandoned.
        !    By default, MIN_RADF = 0.1 * TRUST_RADF, and is also set to this default
        !    value if it is less than DES_TOL. After MIN_RADF and TRUST_RADF are set,
        !    MIN_RADF < TRUST_RADF must hold.
        !
        ! TRUST_RADF defines the initial trust region centered at an isolated
        !    point X as [X - TRUST_RADF * (UB - LB), X + TRUST_RADF * (UB - LB)]
        !    intersected with [LB, UB].  By default, TRUST_RADF = 0.2, and is also set
        !    to this value if the value given is outside the interval
        !    (DES_TOL, 1 - DES_TOL).
        !
        ! PMODE is a logical input that specifies whether or not iteration tasks
        !    should be performed in parallel. By default, PMODE = .FALSE. Note
        !    that this value is not saved during checkpointing, and must be reset
        !    by the user when recovery mode is active, whenever a non-default value
        !    is desired.
        !
        ! In recovery mode the inputs D, P, LB, and UB are still referenced
        !    (for sanity checks). Also, the procedure arguments are still
        !    needed to recover the procedure settings. No other optional
        !    arguments are referenced.
        !"""

        # Setting up "d"
        if type(d) is not ctypes.c_int:
            d = ctypes.c_int(d)

        # Setting up "p"
        if type(p) is not ctypes.c_int:
            p = ctypes.c_int(p)

        # Setting up "lb"
        if ((not issubclass(type(lb), numpy.ndarray)) or (not numpy.asarray(lb).flags.f_contiguous)
                or (not (lb.dtype == numpy.dtype(ctypes.c_double)))):
            import warnings

            warnings.warn("The provided argument 'lb' was not an f_contiguous NumPy array " +
                          "of type 'ctypes.c_double' (or equivalent). Automatically " +
                          "converting (probably creating a full copy).")
            lb = numpy.asarray(lb, dtype=ctypes.c_double, order="F")
        lb_dim_1 = ctypes.c_int(lb.shape[0])

        # Setting up "ub"
        if ((not issubclass(type(ub), numpy.ndarray)) or (not numpy.asarray(ub).flags.f_contiguous)
                or (not (ub.dtype == numpy.dtype(ctypes.c_double)))):
            import warnings

            warnings.warn("The provided argument 'ub' was not an f_contiguous NumPy array " +
                          "of type 'ctypes.c_double' (or equivalent). Automatically " +
                          "converting (probably creating a full copy).")
            ub = numpy.asarray(ub, dtype=ctypes.c_double, order="F")
        ub_dim_1 = ctypes.c_int(ub.shape[0])

        # Setting up "ierr"
        ierr = ctypes.c_int()

        # Setting up "lopt_budget"
        if type(lopt_budget) is not ctypes.c_int:
            lopt_budget = ctypes.c_int(lopt_budget)

        # Setting up "decay"
        if type(decay) is not ctypes.c_double:
            decay = ctypes.c_double(decay)

        # Setting up "des_tol"
        if type(des_tol) is not ctypes.c_double:
            des_tol = ctypes.c_double(des_tol)

        # Setting up "eps"
        if type(eps) is not ctypes.c_double:
            eps = ctypes.c_double(eps)

        # Setting up "epsw"
        if type(epsw) is not ctypes.c_double:
            epsw = ctypes.c_double(epsw)

        # Setting up "obj_tol"
        if type(obj_tol) is not ctypes.c_double:
            obj_tol = ctypes.c_double(obj_tol)

        # Setting up "min_radf"
        if type(min_radf) is not ctypes.c_double:
            min_radf = ctypes.c_double(min_radf)

        # Setting up "trust_radf"
        if type(trust_radf) is not ctypes.c_double:
            trust_radf = ctypes.c_double(trust_radf)

        # Setting up "obj_bounds"
        if ((not issubclass(type(obj_bounds), numpy.ndarray)) or (not numpy.asarray(obj_bounds).flags.f_contiguous)
                or (not (obj_bounds.dtype == numpy.dtype(ctypes.c_double)))):
            import warnings

            warnings.warn("The provided argument 'obj_bounds' was not an f_contiguous " +
                          "NumPy array of type 'ctypes.c_double' (or equivalent). " +
                          "Automatically converting (probably creating a full copy).")
            obj_bounds = numpy.asarray(obj_bounds, dtype=ctypes.c_double, order="F")
        obj_bounds_dim_1 = ctypes.c_int(obj_bounds.shape[0])
        obj_bounds_dim_2 = ctypes.c_int(obj_bounds.shape[1])

        # Setting up "pmode"
        if type(pmode) is not ctypes.c_bool:
            pmode = ctypes.c_bool(pmode)

        # Call C-accessible Fortran wrapper.
        clib.c_vtmop_libe_init(
            ctypes.byref(d),
            ctypes.byref(p),
            ctypes.byref(lb_dim_1),
            ctypes.c_void_p(lb.ctypes.data),
            ctypes.byref(ub_dim_1),
            ctypes.c_void_p(ub.ctypes.data),
            ctypes.byref(ierr),
            ctypes.byref(lopt_budget),
            ctypes.byref(decay),
            ctypes.byref(des_tol),
            ctypes.byref(eps),
            ctypes.byref(epsw),
            ctypes.byref(obj_tol),
            ctypes.byref(min_radf),
            ctypes.byref(trust_radf),
            ctypes.byref(obj_bounds_dim_1),
            ctypes.byref(obj_bounds_dim_2),
            ctypes.c_void_p(obj_bounds.ctypes.data),
            ctypes.byref(pmode),
        )

        # Return final results, 'INTENT(OUT)' arguments only.
        return ierr.value

    # ----------------------------------------------
    # Wrapper for the Fortran subroutine VTMOP_LIBE_GENERATE

    def vtmop_libe_generate(self, d, p, lb, ub, des_pts, obj_pts, isnb, snb, onb, obj_bounds):
        """! The VTMOP_LIBE_GENERATE subroutine produces static batches of candidate
        ! points using VTMOP_MOD. Each call to VTMOP_LIBE_GENERATE performs a half
        ! iteration of the VTMOP_SOLVE algorithm (from VTMOP_LIB), using a Latin
        ! hypercube design of size SNB.
        !
        ! VTMOP_INITIALIZE must still be used to initialize the VTMOP object, and
        ! VTMOP_FINALIZE must be used to terminate and post process the results.
        !
        !
        ! On input:
        !
        ! D is the dimension of the design space.
        !
        ! P is the dimension of the objective space.
        !
        ! LB(1:D) is the real vector of lower bound constraints for the
        !    D design variables.
        !
        ! UB(1:D) is the real vector of upper bound constraints for the
        !    D design variables.
        !
        ! DES_PTS(D,N) is the current list of design points.
        !
        ! OBJ_PTS(P,N) is the current list of objective points.
        !
        ! ISNB is the preferred initial search batch size and is used as the
        !    size of the Latin hypercube design in the zeroeth iteration
        !    (initial search). ISNB should be much greater than SNB.
        !
        ! SNB is the preferred search batch size and is used as the
        !    size of the Latin hypercube design in the search phase.
        !
        ! ONB is the preferred optimization batch size and is used to
        !    pad out the batch of candidate designs in the optimization phase.
        !
        !
        ! On output:
        !
        ! LBATCH returns the size of the next batch.
        !
        ! BATCHX(D,LBATCH) is the next batch of design points to evaluate.
        !
        ! IERR is an integer error flag. Error codes carried from worker
        !    subroutines. In general:
        !
        !  000 : successful iteration.
        !  003 : stopping criterion 3 achieved.
        !
        !  1xx : illegal input error.
        !   Tens digit:
        !    11x : The VTMOP object contained illegal dimensions or values, and
        !          may not have been properly initialized.
        !       Ones digit:
        !         110 : The problem dimensions D and P are not legal.
        !         111 : The internal arrays have not been allocated.
        !         112 : The internal arrays have been initialized, but to invalid
        !               dimensions.
        !         113 : The lower and upper bound constraints contain illegal values,
        !               subject to the design space tolerance.
        !    12x : Illegal values in other inputs.
        !         120 : The lead dimension of DES_PTS does not match the data in VTMOP.
        !         121 : The lead dimension of OBJ_PTS does not match the data in VTMOP.
        !         122 : The second dimensions of DES_PTS and OBJ_PTS do not match.
        !         123 : The preferred batch size NB must be nonnegative.
        !
        !    2xx, 3xx : Error in VTMOP_LTR or VTMOP_OPT, respectively.
        !
        !    5xx, 6xx, 7xx, 8xx : Error in DELAUNAYGRAPH, FIT_SURROGATES, LOCAL_OPT,
        !                         or LH_DESIGN, respectively. See documentation for
        !                         those procedures for more details.
        !
        !    9xx : Checkpointing error. See checkpointing subroutine documentation.
        !"""

        # Setting up "d"
        if type(d) is not ctypes.c_int:
            d = ctypes.c_int(d)

        # Setting up "p"
        if type(p) is not ctypes.c_int:
            p = ctypes.c_int(p)

        # Setting up "lb"
        if ((not issubclass(type(lb), numpy.ndarray)) or (not numpy.asarray(lb).flags.f_contiguous)
                or (not (lb.dtype == numpy.dtype(ctypes.c_double)))):
            import warnings

            warnings.warn("The provided argument 'lb' was not an f_contiguous NumPy array " +
                          "of type 'ctypes.c_double' (or equivalent). Automatically " +
                          "converting (probably creating a full copy).")
            lb = numpy.asarray(lb, dtype=ctypes.c_double, order="F")
        lb_dim_1 = ctypes.c_int(lb.shape[0])

        # Setting up "ub"
        if ((not issubclass(type(ub), numpy.ndarray)) or (not numpy.asarray(ub).flags.f_contiguous)
                or (not (ub.dtype == numpy.dtype(ctypes.c_double)))):
            import warnings

            warnings.warn("The provided argument 'ub' was not an f_contiguous NumPy array " +
                          "of type 'ctypes.c_double' (or equivalent). Automatically " +
                          "converting (probably creating a full copy).")
            ub = numpy.asarray(ub, dtype=ctypes.c_double, order="F")
        ub_dim_1 = ctypes.c_int(ub.shape[0])

        # Setting up "des_pts"
        if ((not issubclass(type(des_pts), numpy.ndarray)) or (not numpy.asarray(des_pts).flags.f_contiguous)
                or (not (des_pts.dtype == numpy.dtype(ctypes.c_double)))):
            import warnings

            warnings.warn("The provided argument 'des_pts' was not an f_contiguous NumPy " +
                          "array of type 'ctypes.c_double' (or equivalent). Automatically " +
                          "converting (probably creating a full copy).")
            des_pts = numpy.asarray(des_pts, dtype=ctypes.c_double, order="F")
        des_pts_dim_1 = ctypes.c_int(des_pts.shape[0])
        des_pts_dim_2 = ctypes.c_int(des_pts.shape[1])

        # Setting up "obj_pts"
        if ((not issubclass(type(obj_pts), numpy.ndarray)) or (not numpy.asarray(obj_pts).flags.f_contiguous)
                or (not (obj_pts.dtype == numpy.dtype(ctypes.c_double)))):
            import warnings

            warnings.warn("The provided argument 'obj_pts' was not an f_contiguous NumPy " +
                          "array of type 'ctypes.c_double' (or equivalent). Automatically " +
                          "converting (probably creating a full copy).")
            obj_pts = numpy.asarray(obj_pts, dtype=ctypes.c_double, order="F")
        obj_pts_dim_1 = ctypes.c_int(obj_pts.shape[0])
        obj_pts_dim_2 = ctypes.c_int(obj_pts.shape[1])

        # Setting up "isnb"
        if type(isnb) is not ctypes.c_int:
            isnb = ctypes.c_int(isnb)

        # Setting up "snb"
        if type(snb) is not ctypes.c_int:
            snb = ctypes.c_int(snb)

        # Setting up "onb"
        if type(onb) is not ctypes.c_int:
            onb = ctypes.c_int(onb)

        # Setting up "obj_bounds"
        if ((not issubclass(type(obj_bounds), numpy.ndarray)) or (not numpy.asarray(obj_bounds).flags.f_contiguous)
                or (not (obj_bounds.dtype == numpy.dtype(ctypes.c_double)))):
            import warnings

            warnings.warn("The provided argument 'obj_bounds' was not an f_contiguous " +
                          "NumPy array of type 'ctypes.c_double' (or equivalent). " +
                          "Automatically converting (probably creating a full copy).")
            obj_bounds = numpy.asarray(obj_bounds, dtype=ctypes.c_double, order="F")
        obj_bounds_dim_1 = ctypes.c_int(obj_bounds.shape[0])
        obj_bounds_dim_2 = ctypes.c_int(obj_bounds.shape[1])

        # Setting up "lbatch"
        lbatch = ctypes.c_int()

        # Setting up "batchx"
        batchx = ctypes.c_void_p()
        batchx_dim_1 = ctypes.c_int()
        batchx_dim_2 = ctypes.c_int()

        # Setting up "ierr"
        ierr = ctypes.c_int()

        # Call C-accessible Fortran wrapper.
        clib.c_vtmop_libe_generate(
            ctypes.byref(d),
            ctypes.byref(p),
            ctypes.byref(lb_dim_1),
            ctypes.c_void_p(lb.ctypes.data),
            ctypes.byref(ub_dim_1),
            ctypes.c_void_p(ub.ctypes.data),
            ctypes.byref(des_pts_dim_1),
            ctypes.byref(des_pts_dim_2),
            ctypes.c_void_p(des_pts.ctypes.data),
            ctypes.byref(obj_pts_dim_1),
            ctypes.byref(obj_pts_dim_2),
            ctypes.c_void_p(obj_pts.ctypes.data),
            ctypes.byref(isnb),
            ctypes.byref(snb),
            ctypes.byref(onb),
            ctypes.byref(obj_bounds_dim_1),
            ctypes.byref(obj_bounds_dim_2),
            ctypes.c_void_p(obj_bounds.ctypes.data),
            ctypes.byref(lbatch),
            ctypes.byref(batchx_dim_1),
            ctypes.byref(batchx_dim_2),
            ctypes.byref(batchx),
            ctypes.byref(ierr),
        )

        # Post-processing "batchx"
        batchx_size = (batchx_dim_1.value) * (batchx_dim_2.value)
        if batchx_size > 0:
            batchx = numpy.array(
                ctypes.cast(batchx, ctypes.POINTER(ctypes.c_double * batchx_size)).contents, copy=False)
            batchx = batchx.reshape(batchx_dim_2.value, batchx_dim_1.value).T
        elif batchx_size == 0:
            batchx = numpy.zeros((batchx_dim_2.value, batchx_dim_1.value), dtype=ctypes.c_double, order="F")
        else:
            batchx = None

        # Return final results, 'INTENT(OUT)' arguments only.
        return lbatch.value, batchx, ierr.value


vtmop_libe_mod = vtmop_libe_mod()



import os
import time
import numpy as np

from libensemble.executors.executor import Executor
from libensemble.message_numbers import WORKER_DONE, TASK_FAILED
from read_sim_output import read_sim_output
from write_sim_input import write_sim_input

"""
This file is part of the suite of scripts to use LibEnsemble on top of WarpX
simulations. It defines a sim_f function that takes LibEnsemble history and
input parameters, run a WarpX simulation and returns 'f'.
"""


def run_warpx(H, persis_info, sim_specs, libE_info):
    """
    This function runs a WarpX simulation and returns quantity 'f' as well as
    other physical quantities measured in the run for convenience. Status check
    is done periodically on the simulation, provided by LibEnsemble.
    """

    # Setting up variables needed for input and output
    # keys              = variable names
    # x                 = variable values
    # libE_output       = what will be returned to libE

    calc_status = 0  # Returns to worker

    input_file = sim_specs['user']['input_filename']
    time_limit = sim_specs['user']['sim_kill_minutes'] * 60.0
    machine_specs = sim_specs['user']['machine_specs']

    exctr = Executor.executor  # Get Executor

    # Modify WarpX input file with input parameters calculated by gen_f
    # and passed to this sim_f.
    write_sim_input(input_file, H['x'])

    # Passed to command line in addition to the executable.
    # Here, only input file
    app_args = input_file
    os.environ["OMP_NUM_THREADS"] = machine_specs['OMP_NUM_THREADS']

    # Launch the executor to actually run the WarpX simulation
    if machine_specs['name'] == 'summit':
        task = exctr.submit(app_name='warpx',
                            extra_args=machine_specs['extra_args'],
                            app_args=app_args,
                            stdout='out.txt',
                            stderr='err.txt',
                            wait_on_start=True)
    else:
        task = exctr.submit(app_name='warpx',
                            num_procs=machine_specs['cores'],
                            app_args=app_args,
                            stdout='out.txt',
                            stderr='err.txt',
                            wait_on_start=True)

    # Periodically check the status of the simulation
    poll_interval = 1  # secs
    while(not task.finished):
        time.sleep(poll_interval)
        task.poll()
        if task.runtime > time_limit:
            task.kill()  # Timeout

    # Set calc_status with optional prints.
    if task.finished:
        if task.state == 'FINISHED':
            calc_status = WORKER_DONE
        elif task.state == 'FAILED':
            print("Warning: Task {} failed: Error code {}"
                  .format(task.name, task.errcode))
            calc_status = TASK_FAILED
        elif task.state == 'USER_KILLED':
            print("Warning: Task {} has been killed"
                  .format(task.name))
        else:
            print("Warning: Task {} in unknown state {}. Error code {}"
                  .format(task.name, task.state, task.errcode))

    # Safety
    time.sleep(0.2)

    try:
        # Get output from a run and delete output files
        warpx_out = read_sim_output(task.workdir)
    except Exception:
        warpx_out = np.nan
        print('Warning - output is Nan')

    # Pass the sim output values to LibEnsemble.
    # When optimization is ON, 'f' is then passed to the generating function
    # gen_f to generate new inputs for next runs.
    # All other parameters are here just for convenience.
    libE_output = np.zeros(1, dtype=sim_specs['out'])

    if np.isnan(warpx_out):
        libE_output['f'] = np.nan
    else:
        libE_output['f'] = warpx_out[0]
        libE_output['energy_std'] = warpx_out[1]
        libE_output['energy_avg'] = warpx_out[2]
        libE_output['charge'] = warpx_out[3]
        libE_output['emittance'] = warpx_out[4]
        libE_output['ramp_down_1'] = H['x'][0][0]
        libE_output['ramp_down_2'] = H['x'][0][1]
        libE_output['zlens_1'] = H['x'][0][2]
        libE_output['adjust_factor'] = H['x'][0][3]

    return libE_output, persis_info, calc_status



"""
This script is submitted as an app and job to Balsam. The job submission is via
'balsam launch' executed in the test_balsam_hworld.py script.
"""

import numpy as np
import mpi4py
from mpi4py import MPI

from libensemble.executors.balsam_executor import BalsamMPIExecutor
from libensemble.message_numbers import WORKER_DONE, WORKER_KILL_ON_ERR, WORKER_KILL_ON_TIMEOUT, TASK_FAILED
from libensemble.libE import libE
from libensemble.sim_funcs.executor_hworld import executor_hworld
from libensemble.gen_funcs.sampling import uniform_random_sample
from libensemble.tools import add_unique_random_streams
import libensemble.sim_funcs.six_hump_camel as six_hump_camel

mpi4py.rc.recv_mprobe = False  # Disable matching probes

libE_specs = {
    'mpi_comm': MPI.COMM_WORLD,
    'comms': 'mpi',
    'save_every_k_sims': 400,
    'save_every_k_gens': 20,
}

nworkers = MPI.COMM_WORLD.Get_size() - 1
is_manager = MPI.COMM_WORLD.Get_rank() == 0

cores_per_task = 1

sim_app = './my_simtask.x'
sim_app2 = six_hump_camel.__file__

exctr = BalsamMPIExecutor()
exctr.register_app(full_path=sim_app, calc_type='sim')  # Default 'sim' app - backward compatible
exctr.register_app(full_path=sim_app2, app_name='six_hump_camel')  # Named app
exctr.register_app(full_path=sim_app2, app_name='sim_hump_camel_dry_run')

sim_specs = {
    'sim_f': executor_hworld,
    'in': ['x'],
    'out': [('f', float), ('cstat', int)],
    'user': {
        'cores': cores_per_task,
        'balsam_test': True,
    },
}

gen_specs = {
    'gen_f': uniform_random_sample,
    'in': ['sim_id'],
    'out': [('x', float, (2,))],
    'user': {
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
        'gen_batch_size': nworkers,
    },
}

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {'elapsed_wallclock_time': 60}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, libE_specs=libE_specs)

if is_manager:
    print('\nChecking expected task status against Workers ...\n')
    calc_status_list_in = np.asarray(
        [WORKER_DONE, WORKER_KILL_ON_ERR, WORKER_DONE, WORKER_KILL_ON_TIMEOUT, TASK_FAILED, 0]
    )
    calc_status_list = np.repeat(calc_status_list_in, nworkers)

    print("Expecting: {}".format(calc_status_list))
    print("Received:  {}\n".format(H['cstat']))

    assert np.array_equal(H['cstat'], calc_status_list), "Error - unexpected calc status. Received: " + str(H['cstat'])

    # Check dry_run submissions inside ensemble.log
    with open('ensemble.log', 'r') as f:
        lines = f.readlines()

    test = len([i for i in lines if 'Test (No submit) Runline:' in i]) == (len(calc_status_list_in) - 1) * nworkers
    assert test, "Dry run runlines not listed in ensemble.log for each dry_run submission instance."

    # Cleanup (maybe cover del_apps() and del_tasks())
    exctr.del_apps()
    exctr.del_tasks()

    print("\n\n\nRun completed.")



"""
This module contains a simple calibration example using the Surmise package.
"""

__all__ = ['surmise_calib']

import numpy as np
from libensemble.gen_funcs.surmise_calib_support import gen_xs, gen_thetas, gen_observations, gen_true_theta, \
    thetaprior, select_next_theta
from surmise.calibration import calibrator
from surmise.emulation import emulator
from libensemble.message_numbers import STOP_TAG, PERSIS_STOP, FINISHED_PERSISTENT_GEN_TAG, EVAL_GEN_TAG
from libensemble.tools.persistent_support import PersistentSupport


def build_emulator(theta, x, fevals):
    """Build the emulator."""
    print(x.shape, theta.shape, fevals.shape)
    emu = emulator(x, theta, fevals, method='PCGPwM',
                   options={'xrmnan': 'all',
                            'thetarmnan': 'never',
                            'return_grad': True})
    emu.fit()
    return emu


def select_condition(pending, n_remaining_theta=5):
    n_x = pending.shape[0]
    return False if np.sum(pending) > n_remaining_theta * n_x else True


def rebuild_condition(pending, prev_pending, n_theta=5):  # needs changes
    n_x = pending.shape[0]
    if np.sum(prev_pending) - np.sum(pending) >= n_x * n_theta or np.sum(pending) == 0:
        return True
    else:
        return False


def create_arrays(calc_in, n_thetas, n_x):
    """Create 2D (point * rows) arrays fevals, pending and complete"""
    fevals = np.reshape(calc_in['f'], (n_x, n_thetas))
    pending = np.full(fevals.shape, False)
    prev_pending = pending.copy()
    complete = np.full(fevals.shape, True)

    return fevals, pending, prev_pending, complete


def pad_arrays(n_x, thetanew, theta, fevals, pending, prev_pending, complete):
    """Extend arrays to appropriate sizes."""
    n_thetanew = len(thetanew)

    theta = np.vstack((theta, thetanew))
    fevals = np.hstack((fevals, np.full((n_x, n_thetanew), np.nan)))
    pending = np.hstack((pending, np.full((n_x, n_thetanew), True)))
    prev_pending = np.hstack((prev_pending, np.full((n_x, n_thetanew), True)))
    complete = np.hstack((complete, np.full((n_x, n_thetanew), False)))

    # print('after:', fevals.shape, theta.shape, pending.shape, complete.shape)
    return theta, fevals, pending, prev_pending, complete


def update_arrays(fevals, pending, complete, calc_in, obs_offset, n_x):
    """Unpack from calc_in into 2D (point * rows) fevals"""
    sim_id = calc_in['sim_id']
    c, r = divmod(sim_id - obs_offset, n_x)  # r, c are arrays if sim_id is an array

    fevals[r, c] = calc_in['f']
    pending[r, c] = False
    complete[r, c] = True
    return


def cancel_columns(obs_offset, c, n_x, pending, ps):
    """Cancel columns"""
    sim_ids_to_cancel = []
    columns = np.unique(c)
    for c in columns:
        col_offset = c*n_x
        for i in range(n_x):
            sim_id_cancl = obs_offset + col_offset + i
            if pending[i, c]:
                sim_ids_to_cancel.append(sim_id_cancl)
                pending[i, c] = 0

    # Send only these fields to existing H rows and libEnsemble will slot in the change.
    H_o = np.zeros(len(sim_ids_to_cancel), dtype=[('sim_id', int), ('cancel_requested', bool)])
    H_o['sim_id'] = sim_ids_to_cancel
    H_o['cancel_requested'] = True
    ps.send(H_o)


def assign_priority(n_x, n_thetas):
    """Assign priorities to points."""
    # Arbitrary priorities
    priority = np.arange(n_x*n_thetas)
    np.random.shuffle(priority)
    return priority


def load_H(H, xs, thetas, offset=0, set_priorities=False):
    """Fill inputs into H0.

    There will be num_points x num_thetas entries
    """
    n_thetas = len(thetas)
    for i, x in enumerate(xs):
        start = (i+offset)*n_thetas
        H['x'][start:start+n_thetas] = x
        H['thetas'][start:start+n_thetas] = thetas

    if set_priorities:
        n_x = len(xs)
        H['priority'] = assign_priority(n_x, n_thetas)


def gen_truevals(x, gen_specs):
    """Generate true values using libE."""
    n_x = len(x)
    H_o = np.zeros((1) * n_x, dtype=gen_specs['out'])

    # Generate true theta and load into H
    true_theta = gen_true_theta()
    H_o['x'][0:n_x] = x
    H_o['thetas'][0:n_x] = true_theta
    return H_o


def surmise_calib(H, persis_info, gen_specs, libE_info):
    """Generator to select and obviate parameters for calibration."""
    rand_stream = persis_info['rand_stream']
    n_thetas = gen_specs['user']['n_init_thetas']
    n_x = gen_specs['user']['num_x_vals']  # Num of x points
    step_add_theta = gen_specs['user']['step_add_theta']  # No. of thetas to generate per step
    n_explore_theta = gen_specs['user']['n_explore_theta']  # No. of thetas to explore
    obsvar_const = gen_specs['user']['obsvar']  # Constant for generator
    priorloc = gen_specs['user']['priorloc']
    priorscale = gen_specs['user']['priorscale']
    ps = PersistentSupport(libE_info, EVAL_GEN_TAG)

    prior = thetaprior(priorloc, priorscale)

    # Create points at which to evaluate the sim
    x = gen_xs(n_x, rand_stream)

    H_o = gen_truevals(x, gen_specs)
    obs_offset = len(H_o)

    tag, Work, calc_in = ps.send_recv(H_o)
    if tag in [STOP_TAG, PERSIS_STOP]:
        return H, persis_info, FINISHED_PERSISTENT_GEN_TAG

    returned_fevals = np.reshape(calc_in['f'], (1, n_x))
    true_fevals = returned_fevals
    obs, obsvar = gen_observations(true_fevals, obsvar_const, rand_stream)

    # Generate a batch of inputs and load into H
    H_o = np.zeros(n_x*(n_thetas), dtype=gen_specs['out'])
    theta = gen_thetas(prior, n_thetas)
    load_H(H_o, x, theta, set_priorities=True)
    tag, Work, calc_in = ps.send_recv(H_o)
    # -------------------------------------------------------------------------

    fevals = None
    prev_pending = None

    while tag not in [STOP_TAG, PERSIS_STOP]:
        if fevals is None:  # initial batch
            fevals, pending, prev_pending, complete = create_arrays(calc_in, n_thetas, n_x)
            emu = build_emulator(theta, x, fevals)
            # Refer to surmise package for additional options
            cal = calibrator(emu, obs, x, prior, obsvar, method='directbayes')

            print('quantiles:', np.round(np.quantile(cal.theta.rnd(10000), (0.01, 0.99), axis=0), 3))
            update_model = False
        else:
            # Update fevals from calc_in
            update_arrays(fevals, pending, complete, calc_in,
                          obs_offset, n_x)
            update_model = rebuild_condition(pending, prev_pending)
            if not update_model:
                tag, Work, calc_in = ps.recv()
                if tag in [STOP_TAG, PERSIS_STOP]:
                    break

        if update_model:
            print('Percentage Cancelled: %0.2f ( %d / %d)' % (100*np.round(np.mean(1-pending-complete), 4),
                                                              np.sum(1-pending-complete),
                                                              np.prod(pending.shape)))
            print('Percentage Pending: %0.2f ( %d / %d)' % (100*np.round(np.mean(pending), 4),
                                                            np.sum(pending),
                                                            np.prod(pending.shape)))
            print('Percentage Complete: %0.2f ( %d / %d)' % (100*np.round(np.mean(complete), 4),
                                                             np.sum(complete),
                                                             np.prod(pending.shape)))

            emu.update(theta=theta, f=fevals)
            cal.fit()

            samples = cal.theta.rnd(2500)
            print(np.mean(np.sum((samples - np.array([0.5]*4))**2, 1)))
            print(np.round(np.quantile(cal.theta.rnd(10000), (0.01, 0.99), axis=0), 3))

            step_add_theta += 2
            prev_pending = pending.copy()
            update_model = False

        # Conditionally generate new thetas from model
        if select_condition(pending):
            new_theta, info = select_next_theta(step_add_theta, cal, emu, pending, n_explore_theta)

            # Add space for new thetas
            theta, fevals, pending, prev_pending, complete = \
                pad_arrays(n_x, new_theta, theta, fevals, pending, prev_pending, complete)

            # n_thetas = step_add_theta
            H_o = np.zeros(n_x*(len(new_theta)), dtype=gen_specs['out'])
            load_H(H_o, x, new_theta, set_priorities=True)
            tag, Work, calc_in = ps.send_recv(H_o)

            # Determine evaluations to cancel
            c_obviate = info['obviatesugg']
            if len(c_obviate) > 0:
                print('columns sent for cancel is:  {}'.format(c_obviate), flush=True)
                cancel_columns(obs_offset, c_obviate, n_x, pending, ps)
            pending[:, c_obviate] = False

    return None, persis_info, FINISHED_PERSISTENT_GEN_TAG



import os
import time
import numpy as np

from libensemble.executors.executor import Executor
from libensemble.message_numbers import WORKER_DONE, WORKER_KILL, TASK_FAILED

MAX_SEED = 32767


class ForcesException(Exception):
    """Custom forces exception"""


def perturb(particles, seed, max_fraction):
    """Modify particle count"""
    seed_fraction = seed/MAX_SEED
    max_delta = particles * max_fraction
    delta = seed_fraction * max_delta
    delta = delta - max_delta/2  # translate so -/+
    new_particles = particles + delta
    return int(new_particles)


def read_last_line(filepath):
    """Read last line of statfile"""
    try:
        with open(filepath, 'rb') as fh:
            line = fh.readlines()[-1].decode().rstrip()
    except Exception:
        line = ""  # In case file is empty or not yet created
    return line


def make_unique_simdir(simdir, count=0):
    """As some dir names could recur, make sure unique"""
    if not os.path.isdir(simdir):
        return simdir
    else:
        count += 1
        return make_unique_simdir(".".join([simdir.split('.')[0], str(count)]), count)


def run_forces(H, persis_info, sim_specs, libE_info):
    # Setting up variables needed for input and output
    # keys              = variable names
    # x                 = variable values
    # output            = what will be returned to libE
    if sim_specs['user']['fail_on_sim']:
        raise ForcesException

    calc_status = 0  # Returns to worker

    x = H['x']
    # keys = sim_specs['user']['keys']
    sim_particles = sim_specs['user']['sim_particles']
    sim_timesteps = sim_specs['user']['sim_timesteps']
    time_limit = sim_specs['user']['sim_kill_minutes'] * 60.0

    # Get from dictionary if key exists, else return default (e.g. 0)
    cores = sim_specs['user'].get('cores', None)
    kill_rate = sim_specs['user'].get('kill_rate', 0)
    particle_variance = sim_specs['user'].get('particle_variance', 0)

    # Composing variable names and x values to set up simulation
    seed = int(np.rint(x[0][0]))

    # This is to give a random variance of work-load
    sim_particles = perturb(sim_particles, seed, particle_variance)
    print('seed: {}   particles: {}'.format(seed, sim_particles))

    exctr = Executor.executor  # Get Executor

    args = str(int(sim_particles)) + ' ' + str(sim_timesteps) + ' ' + str(seed) + ' ' + str(kill_rate)
    # task = exctr.submit( app_name='forces', num_procs=cores, app_args=args, stdout='out.txt', stderr='err.txt')

    machinefile = None
    if sim_specs['user']['fail_on_submit']:
        machinefile = 'fail'

    # Machinefile only used here for exception testing
    if cores:
        task = exctr.submit(app_name='forces', num_procs=cores, app_args=args,
                            stdout='out.txt', stderr='err.txt', wait_on_start=True,
                            machinefile=machinefile)
    else:
        task = exctr.submit(app_name='forces', app_args=args, stdout='out.txt',
                            stderr='err.txt', wait_on_start=True, hyperthreads=True,
                            machinefile=machinefile)  # Auto-partition

    # Stat file to check for bad runs
    statfile = 'forces.stat'
    filepath = os.path.join(task.workdir, statfile)
    line = None

    poll_interval = 1  # secs
    while(not task.finished):
        # Read last line of statfile
        line = read_last_line(filepath)
        if line == "kill":
            task.kill()  # Bad run
        elif task.runtime > time_limit:
            task.kill()  # Timeout
        else:
            time.sleep(poll_interval)
            task.poll()

    if task.finished:
        if task.state == 'FINISHED':
            print("Task {} completed".format(task.name))
            calc_status = WORKER_DONE
            if read_last_line(filepath) == "kill":
                # Generally mark as complete if want results (completed after poll - before readline)
                print("Warning: Task completed although marked as a bad run (kill flag set in forces.stat)")
        elif task.state == 'FAILED':
            print("Warning: Task {} failed: Error code {}".format(task.name, task.errcode))
            calc_status = TASK_FAILED
        elif task.state == 'USER_KILLED':
            print("Warning: Task {} has been killed".format(task.name))
            calc_status = WORKER_KILL
        else:
            print("Warning: Task {} in unknown state {}. Error code {}".format(task.name, task.state, task.errcode))

    time.sleep(0.2)
    try:
        data = np.loadtxt(filepath)
        # task.read_file_in_workdir(statfile)
        final_energy = data[-1]
    except Exception:
        final_energy = np.nan
        # print('Warning - Energy Nan')

    outspecs = sim_specs['out']
    output = np.zeros(1, dtype=outspecs)
    output['energy'][0] = final_energy

    return output, persis_info, calc_status



"""
Runs libEnsemble with uniform random sampling and writes results into sim dirs.
Tests sim_input_dir capabilities

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_sim_input_dir_option.py
   python3 test_sim_input_dir_option.py --nworkers 3 --comms local
   python3 test_sim_input_dir_option.py --nworkers 3 --comms tcp

The number of concurrent evaluations of the objective function will be 4-1=3.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 2 4

import numpy as np
import os

from libensemble.libE import libE
from libensemble.tests.regression_tests.support import write_sim_func as sim_f
from libensemble.gen_funcs.sampling import uniform_random_sample as gen_f
from libensemble.tools import parse_args, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()

sim_input_dir = './sim_input_dir'
dir_to_copy = sim_input_dir + '/copy_this'
o_ensemble = './ensemble_inputdir_w' + str(nworkers) + '_' + libE_specs.get('comms')

for dir in [sim_input_dir, dir_to_copy]:
    if not os.path.isdir(dir):
        os.makedirs(dir, exist_ok=True)

libE_specs['sim_input_dir'] = sim_input_dir
libE_specs['ensemble_dir_path'] = o_ensemble
libE_specs['sim_dirs_make'] = False
libE_specs['sim_dir_symlink_files'] = ['./test_sim_input_dir_option.py']  # to cover FileExistsError catch
libE_specs['ensemble_copy_back'] = True

sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
}

gen_specs = {
    'gen_f': gen_f,
    'out': [('x', float, (1,))],
    'user': {
        'gen_batch_size': 20,
        'lb': np.array([-3]),
        'ub': np.array([3]),
    },
}

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {'sim_max': 21}

H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, libE_specs=libE_specs)

if is_manager:
    assert os.path.isdir(o_ensemble), 'Ensemble directory {} not created.'.format(o_ensemble)
    assert os.path.basename(dir_to_copy) in os.listdir(o_ensemble), 'Input file not copied over.'
    with open(os.path.join(o_ensemble, 'test_sim_out.txt'), 'r') as f:
        lines = f.readlines()

    assert len(lines) == exit_criteria['sim_max'], 'Sim output not written to ensemble dir for each sim call'



import os

"""
This file is part of the suite of scripts to use LibEnsemble on top of WarpX
simulations. It contains a dictionary for machine-specific elements.
"""


local_specs = {
    'name': 'local',  # Machine name
    'cores': 1,  # Number of cores per simulation
    'sim_app': os.environ['HOME'] + '/warpx/Bin/main2d.gnu.TPROF.MPI.OMP.ex',
    'extra_args': '',  # extra arguments passed to mpirun/mpiexec at execution
    'OMP_NUM_THREADS': '2',
    'sim_max': 3  # Maximum number of simulations
}


summit_specs = {
    'name': 'summit',  # Machine name
    'sim_app': os.environ['HOME'] + '/warpx/Bin/main2d.gnu.TPROF.MPI.CUDA.ex',
    # extra arguments passed to jsrun at execution
    'extra_args': '-n 1 -a 1 -g 1 -c 1 --bind=packed:1 --smpiargs="-gpu"',
    'OMP_NUM_THREADS': '1',
    'sim_max': 400  # Maximum number of simulations
}



import numpy as np
from libensemble.message_numbers import EVAL_SIM_TAG, EVAL_GEN_TAG
from libensemble.tools.alloc_support import AllocSupport, InsufficientFreeResources


def only_persistent_gens(W, H, sim_specs, gen_specs, alloc_specs, persis_info, libE_info):
    """
    This allocation function will give simulation work if possible, but
    otherwise start up to ``alloc_specs['user']['num_active_gens']``
    persistent generators (defaulting to one).

    By default, evaluation results are given back to the generator once
    all generated points have been returned from the simulation evaluation.
    If ``alloc_specs['user']['async_return']`` is set to True, then any
    returned points are given back to the generator.

    If any workers are marked as zero_resource_workers, then these will only
    be used for generators.

    If any of the persistent generators has exited, then ensemble shutdown
    is triggered.

    **User options**:

    To be provided in calling script: E.g., ``alloc_specs['user']['async_return'] = True``

    init_sample_size: int, optional
        Initial sample size - always return in batch. Default: 0

    num_active_gens: int, optional
        Maximum number of persistent generators to start. Default: 1

    async_return: boolean, optional
        Return results to gen as they come in (after sample). Default: False (batch return).

    active_recv_gen: boolean, optional
        Create gen in active receive mode. If True, the manager does not need to wait
        for a return from the generator before sending further returned points.
        Default: False


    .. seealso::
        `test_persistent_uniform_sampling.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_persistent_uniform_sampling.py>`_ # noqa
        `test_persistent_uniform_sampling_async.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_persistent_uniform_sampling_async.py>`_ # noqa
        `test_persistent_surmise_calib.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_persistent_surmise_calib.py>`_ # noqa
        `test_persistent_uniform_gen_decides_stop.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_persistent_uniform_gen_decides_stop.py>`_ # noqa
    """

    if libE_info['sim_max_given'] or not libE_info['any_idle_workers']:
        return {}, persis_info

    # Initialize alloc_specs['user'] as user.
    user = alloc_specs.get('user', {})
    sched_opts = user.get('scheduler_opts', {})
    manage_resources = 'resource_sets' in H.dtype.names or libE_info['use_resource_sets']
    active_recv_gen = user.get('active_recv_gen', False)  # Persistent gen can handle irregular communications
    init_sample_size = user.get('init_sample_size', 0)  # Always batch return until this many evals complete
    batch_give = user.get('give_all_with_same_priority', False)

    support = AllocSupport(W, manage_resources, persis_info, sched_opts)
    gen_count = support.count_persis_gens()
    Work = {}

    # Asynchronous return to generator
    async_return = user.get('async_return', False) and sum(H['returned']) >= init_sample_size

    if gen_count < persis_info.get('num_gens_started', 0):
        # When a persistent worker is done, trigger a shutdown (returning exit condition of 1)
        return Work, persis_info, 1

    # Give evaluated results back to a running persistent gen
    for wid in support.avail_worker_ids(persistent=EVAL_GEN_TAG, active_recv=active_recv_gen):
        gen_inds = (H['gen_worker'] == wid)
        returned_but_not_given = np.logical_and.reduce((H['returned'], ~H['given_back'], gen_inds))
        if np.any(returned_but_not_given):
            if async_return or support.all_returned(H, gen_inds):
                point_ids = np.where(returned_but_not_given)[0]
                Work[wid] = support.gen_work(wid, gen_specs['persis_in'], point_ids, persis_info.get(wid),
                                             persistent=True, active_recv=active_recv_gen)
                returned_but_not_given[point_ids] = False

    # Now the give_sim_work_first part
    points_to_evaluate = ~H['given'] & ~H['cancel_requested']
    avail_workers = support.avail_worker_ids(persistent=False, zero_resource_workers=False)
    for wid in avail_workers:

        if not np.any(points_to_evaluate):
            break

        sim_ids_to_send = support.points_by_priority(H, points_avail=points_to_evaluate, batch=batch_give)
        try:
            Work[wid] = support.sim_work(wid, H, sim_specs['in'], sim_ids_to_send, persis_info.get(wid))
        except InsufficientFreeResources:
            break

        points_to_evaluate[sim_ids_to_send] = False

    # Start persistent gens if no worker to give out. Uses zero_resource_workers if defined.
    if not np.any(points_to_evaluate):
        avail_workers = support.avail_worker_ids(persistent=False, zero_resource_workers=True)

        for wid in avail_workers:
            if gen_count < user.get('num_active_gens', 1):
                # Finally, start a persistent generator as there is nothing else to do.
                try:
                    Work[wid] = support.gen_work(wid, gen_specs.get('in', []), range(len(H)), persis_info.get(wid),
                                                 persistent=True, active_recv=active_recv_gen)
                except InsufficientFreeResources:
                    break
                persis_info['num_gens_started'] = persis_info.get('num_gens_started', 0) + 1
                gen_count += 1

    return Work, persis_info, 0


def only_persistent_workers(W, H, sim_specs, gen_specs, alloc_specs, persis_info, libE_info):
    """
    SH: TODO - Need to update docstring...
    This allocation function will give simulation work if possible, but
    otherwise start up to ``alloc_specs['user']['num_active_gens']``
    persistent generators (defaulting to one).

    By default, evaluation results are given back to the generator once
    all generated points have been returned from the simulation evaluation.
    If ``alloc_specs['user']['async_return']`` is set to True, then any
    returned points are given back to the generator.

    If any workers are marked as zero_resource_workers, then these will only
    be used for generators.

    If any of the persistent generators has exited, then ensemble shutdown
    is triggered.

    **User options**:

    To be provided in calling script: E.g., ``alloc_specs['user']['async_return'] = True``

    init_sample_size: int, optional
        Initial sample size - always return in batch. Default: 0

    num_active_gens: int, optional
        Maximum number of persistent generators to start. Default: 1

    async_return: boolean, optional
        Return results to gen as they come in (after sample). Default: False (batch return).

    active_recv_gen: boolean, optional
        Create gen in active receive mode. If True, the manager does not need to wait
        for a return from the generator before sending further returned points.
        Default: False


    .. seealso::
        `test_persistent_gensim_uniform_sampling.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_persistent_gensim_uniform_sampling.py>`_ # noqa
    """

    if libE_info['sim_max_given'] or not libE_info['any_idle_workers']:
        return {}, persis_info

    # Initialize alloc_specs['user'] as user.
    user = alloc_specs.get('user', {})
    sched_opts = user.get('scheduler_opts', {})
    manage_resources = 'resource_sets' in H.dtype.names or libE_info['use_resource_sets']
    active_recv_gen = user.get('active_recv_gen', False)  # Persistent gen can handle irregular communications
    init_sample_size = user.get('init_sample_size', 0)  # Always batch return until this many evals complete
    batch_give = user.get('give_all_with_same_priority', False)

    support = AllocSupport(W, manage_resources, persis_info, sched_opts)
    gen_count = support.count_persis_gens()
    Work = {}

    # Asynchronous return to generator
    async_return = user.get('async_return', False) and sum(H['returned']) >= init_sample_size

    if gen_count < persis_info.get('num_gens_started', 0):
        # When a persistent gen worker is done, trigger a shutdown (returning exit condition of 1)
        return Work, persis_info, 1

    # Give evaluated results back to a running persistent gen
    for wid in support.avail_worker_ids(persistent=EVAL_GEN_TAG, active_recv=active_recv_gen):
        gen_inds = (H['gen_worker'] == wid)
        returned_but_not_given = np.logical_and.reduce((H['returned'], ~H['given_back'], gen_inds))
        if np.any(returned_but_not_given):
            if async_return or support.all_returned(H, gen_inds):
                point_ids = np.where(returned_but_not_given)[0]
                Work[wid] = support.gen_work(wid, gen_specs['persis_in'], point_ids, persis_info.get(wid),
                                             persistent=True, active_recv=active_recv_gen)
                returned_but_not_given[point_ids] = False

    # SH Make sure once set up persistent sim that it W array is correct.

    # Now the give_sim_work_first part
    points_to_evaluate = ~H['given'] & ~H['cancel_requested']
    avail_workers = list(
        set(support.avail_worker_ids(persistent=False, zero_resource_workers=False))
        | set(support.avail_worker_ids(persistent=EVAL_SIM_TAG, zero_resource_workers=False)))
    for wid in avail_workers:

        if not np.any(points_to_evaluate):
            break

        sim_ids_to_send = support.points_by_priority(H, points_avail=points_to_evaluate, batch=batch_give)
        try:
            # Note that resources will not change if worker is already persistent.
            Work[wid] = support.sim_work(wid, H, sim_specs['in'], sim_ids_to_send, persis_info.get(wid),
                                         persistent=True)
        except InsufficientFreeResources:
            break

        points_to_evaluate[sim_ids_to_send] = False

    # Start persistent gens if no worker to give out. Uses zero_resource_workers if defined.
    if not np.any(points_to_evaluate):
        avail_workers = support.avail_worker_ids(persistent=False, zero_resource_workers=True)

        for wid in avail_workers:
            if gen_count < user.get('num_active_gens', 1):
                # Finally, start a persistent generator as there is nothing else to do.
                try:
                    Work[wid] = support.gen_work(wid, gen_specs.get('in', []), range(len(H)), persis_info.get(wid),
                                                 persistent=True, active_recv=active_recv_gen)
                except InsufficientFreeResources:
                    break
                persis_info['num_gens_started'] = persis_info.get('num_gens_started', 0) + 1
                gen_count += 1
    del support
    return Work, persis_info, 0



import time
import sys
import signal


def ignore_handler(signum, frame):
    print("Ignoring SIGTERM")


def main(ignore_term=False, wait_time=-1):
    print("Call with {}, {}".format(ignore_term, wait_time))
    if ignore_term:
        signal.signal(signal.SIGTERM, ignore_handler)
    if wait_time > 0:
        time.sleep(wait_time)
    else:
        while True:
            pass


if __name__ == "__main__":
    main(*[float(x) for x in sys.argv[1:]])



import os
import shutil
import numpy as np
import yt
import scipy.constants as scc
import glob

"""
This file is part of the suite of scripts to use LibEnsemble on top of WarpX
simulations. It reads the output plotfiles of a simulation and returns beam
quantity relevant for the LibEnsemble optimizations ('f') as well as other
beam quantities for convenience.
"""

yt.funcs.mylog.setLevel(50)


def _beam_properties(filepath):
    """
    Reads plotfile filepath and compute and return beam parameters

    Parameters
    ----------
    filepath : path to plotfile to read
    """

    # Read beam quantities from plotfile
    ds = yt.load(filepath)
    ad = ds.all_data()
    w = ad['beam', 'particle_weight'].v
    x = ad['beam', 'particle_position_x'].v
    ux = ad['beam', 'particle_momentum_x'].v/scc.m_e/scc.c
    uy = ad['beam', 'particle_momentum_y'].v/scc.m_e/scc.c
    uz = ad['beam', 'particle_momentum_z'].v/scc.m_e/scc.c

    # Compute beam parameters
    # Defined like that, the beam charge is > 0.
    charge = np.sum(w) * scc.e
    gamma = np.sqrt(1. + ux**2 + uy**2 + uz**2)
    energy_MeV = .511 * (gamma - 1.)
    energy_avg = np.mean(energy_MeV)
    energy_std = np.std(energy_MeV) / energy_avg
    emittance = np.sqrt(np.std(x)**2 * np.std(ux)**2 - np.mean(x*ux)**2)
    return charge, energy_avg, energy_std, emittance


def read_sim_output(workdir):
    """
    Return optimizing quantity 'f' and other parameters for convenience.

    Parameters
    ----------
    workdir : Path to directory where the simulation ran.
    """
    # Get beam properties at the beginning of the run
    datafile = 'diags/plotfiles/plt00000/'
    filepath = os.path.join(workdir, datafile)
    charge_i, _, _, emittance_i = _beam_properties(filepath)

    # Get beam properties at the end of the run
    file_list = glob.glob('diags/plotfiles/plt?????')
    file_list.sort()
    datafile = file_list[-1]
    filepath = os.path.join(workdir, datafile)
    charge_f, energy_avg, energy_std, emittance_f = _beam_properties(filepath)

    # delete simulation results, just to have smaller data
    shutil.rmtree('diags')

    # Build a quantity to minimize (f) that encompasses
    # emittance AND charge loss 1% charge loss has the
    # same impact as doubling the initial emittance.
    # we minimize f!
    f = emittance_f + emittance_i*(1.-charge_f/charge_i)*100
    warpx_out = np.array([f, energy_std, energy_avg, charge_f, emittance_f])

    return warpx_out



# --- Tags

UNSET_TAG = 0

# When received by a worker, tells worker to do a sim eval;
# When received by the manager, tells manager that worker is done with sim eval.
EVAL_SIM_TAG = 1

# When received by a worker, tells worker to do a gen eval;
# When received by the manager, tells manager that worker is done with sim eval.
EVAL_GEN_TAG = 2

STOP_TAG = 3             # Manager tells worker (or persistent calc) to stop
PERSIS_STOP = 4          # Manager tells persistent calculation to stop

# last_message_number_rst_tag

calc_type_strings = {
    EVAL_SIM_TAG: 'sim',
    EVAL_GEN_TAG: 'gen',
    PERSIS_STOP: 'STOP with work',
    None: 'No type set'
}


# --- Signal flags (in message body vs tags)

# first_calc_status_rst_tag
# CALC STATUS/SIGNAL FLAGS
FINISHED_PERSISTENT_SIM_TAG = 11  # tells manager sim_f done persistent mode
FINISHED_PERSISTENT_GEN_TAG = 12  # tells manager gen_f done persistent mode
MAN_SIGNAL_FINISH = 20            # Kill tasks and shutdown worker
MAN_SIGNAL_KILL = 21              # Kill running task - but don't stop worker
WORKER_KILL = 30                  # Worker kills not covered by a more specific case
WORKER_KILL_ON_ERR = 31           # Worker killed due to an error in results
WORKER_KILL_ON_TIMEOUT = 32       # Worker killed on timeout
TASK_FAILED = 33                  # Calc had tasks that failed
WORKER_DONE = 34                  # Calculation was successful
# last_calc_status_rst_tag
CALC_EXCEPTION = 35               # Reserved: Automatically used if gen_f or sim_f raised an exception.

calc_status_strings = {
    FINISHED_PERSISTENT_SIM_TAG: "Persis sim finished",
    FINISHED_PERSISTENT_GEN_TAG: "Persis gen finished",
    MAN_SIGNAL_FINISH: "Manager killed on finish",
    MAN_SIGNAL_KILL: "Manager killed task",
    WORKER_KILL_ON_ERR: "Worker killed task on Error",
    WORKER_KILL_ON_TIMEOUT: "Worker killed task on Timeout",
    WORKER_KILL: "Worker killed",
    TASK_FAILED: "Task Failed",
    WORKER_DONE: "Completed",
    CALC_EXCEPTION: "Exception occurred",
    None: "Unknown Status"
}
# last_calc_status_string_rst_tag



"""
Runs libEnsemble on the 6-hump camel problem. Documented here:
   https://www.sfu.ca/~ssurjano/camel6.html

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_persistent_sim_uniform_sampling.py
   python3 test_persistent_sim_uniform_sampling.py --nworkers 3 --comms local
   python3 test_persistent_sim_uniform_sampling.py --nworkers 3 --comms tcp

When running with the above command, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 3 4

import sys
import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.six_hump_camel import persistent_six_hump_camel as sim_f
from libensemble.gen_funcs.persistent_uniform_sampling import persistent_uniform as gen_f
from libensemble.alloc_funcs.start_only_persistent import only_persistent_workers as alloc_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams

# from libensemble import logger
# logger.set_level('DEBUG')

nworkers, is_manager, libE_specs, _ = parse_args()
libE_specs['num_resource_sets'] = nworkers - 1  # Only matters if sims use resources.

# Only used to test returning/overwriting a point at the end of the persistent sim.
libE_specs['use_persis_return_sim'] = True

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")

n = 2

sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'user': {'replace_final_fields': True},
    'out': [('f', float), ('grad', float, n)],
}

gen_specs = {
    'gen_f': gen_f,
    'in': [],
    'persis_in': ['sim_id', 'f', 'grad'],
    'out': [('x', float, (n,))],
    'user': {
        'initial_batch_size': 5,
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
    },
}

alloc_specs = {'alloc_f': alloc_f}

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {'sim_max': 40, 'elapsed_wallclock_time': 300}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

if is_manager:
    assert len(np.unique(H['gen_time'])) == 8
    assert not any((H['f'] == 0))
    # Should overwrite the last value (in fact last (nworker-1) values) with f(1,1) = 3.23333333
    assert not np.isclose(H['f'][0], 3.23333333)
    assert np.isclose(H['f'][-1], 3.23333333)
    save_libE_output(H, persis_info, __file__, nworkers)



import numpy as np
import time
from libensemble.tools.alloc_support import AllocSupport, InsufficientFreeResources


def give_sim_work_first(W, H, sim_specs, gen_specs, alloc_specs, persis_info, libE_info):
    """
    Decide what should be given to workers. This allocation function gives any
    available simulation work first, and only when all simulations are
    completed or running does it start (at most ``alloc_specs['user']['num_active_gens']``)
    generator instances.

    Allows for a ``alloc_specs['user']['batch_mode']`` where no generation
    work is given out unless all entries in ``H`` are returned.

    Can give points in highest priority, if ``'priority'`` is a field in ``H``.
    If alloc_specs['user']['give_all_with_same_priority'] is set to True, then
    all points with the same priority value are given as a batch to the sim.

    Workers performing sims will be assigned resources given in H['resource_sets']
    this field exists, else defaulting to one. Workers performing gens are
    assigned resource_sets given by persis_info['gen_resources'] or zero.

    This is the default allocation function if one is not defined.

    .. seealso::
        `test_uniform_sampling.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_uniform_sampling.py>`_ # noqa
    """

    user = alloc_specs.get('user', {})

    if 'cancel_sims_time' in user:
        # Cancel simulations that are taking too long
        rows = np.where(np.logical_and.reduce((H['given'], ~H['returned'], ~H['cancel_requested'])))[0]
        inds = time.time() - H['last_given_time'][rows] > user['cancel_sims_time']
        to_request_cancel = rows[inds]
        for row in to_request_cancel:
            H[row]['cancel_requested'] = True

    if libE_info['sim_max_given'] or not libE_info['any_idle_workers']:
        return {}, persis_info

    # Initialize alloc_specs['user'] as user.
    sched_opts = user.get('scheduler_opts', {})
    batch_give = user.get('give_all_with_same_priority', False)
    gen_in = gen_specs.get('in', [])

    manage_resources = 'resource_sets' in H.dtype.names or libE_info['use_resource_sets']
    support = AllocSupport(W, manage_resources, persis_info, sched_opts)
    gen_count = support.count_gens()
    Work = {}

    points_to_evaluate = ~H['given'] & ~H['cancel_requested']
    for wid in support.avail_worker_ids():

        if np.any(points_to_evaluate):
            sim_ids_to_send = support.points_by_priority(H, points_avail=points_to_evaluate, batch=batch_give)
            try:
                Work[wid] = support.sim_work(wid, H, sim_specs['in'], sim_ids_to_send, persis_info.get(wid))
            except InsufficientFreeResources:
                break
            points_to_evaluate[sim_ids_to_send] = False
        else:

            # Allow at most num_active_gens active generator instances
            if gen_count >= user.get('num_active_gens', gen_count + 1):
                break

            # Do not start gen instances in batch mode if workers still working
            if user.get('batch_mode') and not support.all_returned(H):
                break

            # Give gen work
            return_rows = range(len(H)) if gen_in else []
            try:
                Work[wid] = support.gen_work(wid, gen_in, return_rows, persis_info.get(wid))
            except InsufficientFreeResources:
                break
            gen_count += 1

    return Work, persis_info



"""
Example using NSGA2 as a libE generator function using the libE yaml interface
For more about NSGA2, see
https://gist.github.com/darden1/fa8f96185a46796ed9516993bfe24862
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local
# TESTSUITE_NPROCS: 3 4

import numpy as np
from time import time
import os
from libensemble import Ensemble
from libensemble.sim_funcs.six_hump_camel import six_hump_camel_func


def deap_six_hump(H, persis_info, sim_specs, _):
    xvals = H['individual'][0]
    Out = np.zeros(1, dtype=sim_specs['out'])
    y0 = np.linalg.norm(xvals)
    y1 = six_hump_camel_func(xvals)
    Out['fitness_values'] = (y0, y1)  # Requires tuple, even with 1 objective

    return Out, persis_info


if __name__ == "__main__":

    deap_test = Ensemble()
    deap_test.from_yaml('deap_nsga2.yaml')

    if deap_test.is_manager:
        start_time = time()

    assert deap_test.nworkers >= 2, "Cannot run with a persistent gen_f if only one worker."

    ind_size = 2
    w = (-1.0, -1.0)  # Must be a tuple

    deap_test.gen_specs['user'].update(
        {
            'weights': w,
            'indpb': 0.8 / ind_size,
        }
    )

    lb = deap_test.gen_specs['user']['lb']
    ub = deap_test.gen_specs['user']['ub']

    for run in range(2):

        deap_test.persis_info.add_random_streams()

        if run == 1:

            # Number of points in the sample
            num_samp = 100

            H0_dtype = [
                ('individual', float, ind_size),
                ('generation', int),
                ('fitness_values', float, 2),
                ('sim_id', int),
                ('returned', bool),
                ('given_back', bool),
                ('given', bool),
            ]

            H0 = np.zeros(num_samp, dtype=H0_dtype)

            # Mark these points as already have been given to be evaluated, and returned, but not given_back.
            H0[['given', 'given_back', 'returned']] = True
            H0['generation'][:] = 1
            # Give these points sim_ids
            H0['sim_id'] = range(num_samp)

            # "Load in" the points and their function values. (In this script, we are
            # actually evaluating them, but in many cases, they are available from past
            # evaluations
            np.random.seed(0)
            H0['individual'] = np.random.uniform(lb, ub, (num_samp, len(lb)))
            for i, x in enumerate(H0['individual']):
                H_dummy = np.zeros(1, dtype=[('individual', float, ind_size)])
                H_dummy['individual'] = x
                objs = deap_six_hump(H_dummy, {}, deap_test.sim_specs, {})
                H0['fitness_values'][i] = objs[0]

            # Testing use_persis_return_gen capabilities
            deap_test.libE_specs['use_persis_return_gen'] = True
            deap_test.H0 = H0
        else:
            deap_test.H0 = None

        # Perform the run
        deap_test.run()

        if deap_test.is_manager:
            if run == 0:
                assert np.sum(deap_test.H['last_points']) == 0, (
                    "The last_points shouldn't be marked (even though "
                    "they were marked in the gen) as 'use_persis_return_gen' was false."
                )
            elif run == 1:
                assert np.sum(deap_test.H['last_points']) == 100, (
                    "The last_points should be marked as true because they "
                    "were marked in the manager and 'use_persis_return_gen' is true."
                )

            script_name = os.path.splitext(os.path.basename(__file__))[0]
            assert deap_test.flag == 0, script_name + " didn't exit correctly"
            assert sum(deap_test.H['returned']) >= deap_test.exit_criteria['sim_max'], (
                script_name + " didn't evaluate the sim_max points."
            )
            assert min(deap_test.H['fitness_values'][:, 0]) <= 4e-3, (
                script_name + " didn't find the minimum for objective 0."
            )
            assert min(deap_test.H['fitness_values'][:, 1]) <= -1.0, (
                script_name + " didn't find the minimum for objective 1."
            )



"""
Runs libEnsemble testing the executor functionality.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_executor_hworld.py
   python3 test_executor_hworld.py --nworkers 3 --comms local
   python3 test_executor_hworld.py --nworkers 3 --comms tcp

The number of concurrent evaluations of the objective function will be 4-1=3.
"""

import os
import numpy as np
import multiprocessing

# Import libEnsemble items for this test
# from libensemble.calc_info import CalcInfo
# from libensemble.executors.executor import Executor
# from libensemble.resources.resources import Resources
from libensemble.message_numbers import WORKER_DONE, WORKER_KILL_ON_ERR, WORKER_KILL_ON_TIMEOUT, TASK_FAILED
from libensemble.libE import libE
from libensemble.sim_funcs.executor_hworld import executor_hworld as sim_f
import libensemble.sim_funcs.six_hump_camel as six_hump_camel
from libensemble.gen_funcs.sampling import uniform_random_sample as gen_f
from libensemble.tools import parse_args, add_unique_random_streams
from libensemble.tests.regression_tests.common import build_simfunc

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 2 3 4

nworkers, is_manager, libE_specs, _ = parse_args()

libE_specs['disable_resource_manager'] = True

USE_BALSAM = False

cores_per_task = 1
logical_cores = multiprocessing.cpu_count()
cores_all_tasks = nworkers * cores_per_task

if cores_all_tasks > logical_cores:
    disable_resource_manager = True
    mess_resources = 'Oversubscribing - Resource manager disabled'
elif libE_specs.get('comms', False) == 'tcp':
    disable_resource_manager = True
    mess_resources = 'TCP comms does not support resource management. Resource manager disabled'
else:
    disable_resource_manager = False
    mess_resources = 'Resource manager enabled'

if is_manager:
    print('\nCores req: {} Cores avail: {}\n  {}\n'.format(cores_all_tasks, logical_cores, mess_resources))

sim_app = './my_simtask.x'
if not os.path.isfile(sim_app):
    build_simfunc()
sim_app2 = six_hump_camel.__file__

if USE_BALSAM:
    from libensemble.executors.balsam_executor import BalsamMPIExecutor

    exctr = BalsamMPIExecutor()
else:
    from libensemble.executors.mpi_executor import MPIExecutor

    exctr = MPIExecutor()
exctr.register_app(full_path=sim_app, calc_type='sim')  # Default 'sim' app - backward compatible
exctr.register_app(full_path=sim_app2, app_name='six_hump_camel')  # Named app

# if nworkers == 3:
#    CalcInfo.keep_worker_stat_files = True # Testing this functionality
# else:
#    CalcInfo.keep_worker_stat_files = False # Testing this functionality

sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float), ('cstat', int)],
    'user': {'cores': cores_per_task},
}

gen_specs = {
    'gen_f': gen_f,
    'in': ['sim_id'],
    'out': [('x', float, (2,))],
    'user': {
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
        'gen_batch_size': nworkers,
    },
}

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {'elapsed_wallclock_time': 30}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, libE_specs=libE_specs)

if is_manager:
    print('\nChecking expected task status against Workers ...\n')

    # Expected list: Last is zero as will not be entered into H array on
    # manager kill - but should show in the summary file.
    # Repeat expected lists nworkers times and compare with list of status's
    # received from workers
    calc_status_list_in = np.asarray(
        [WORKER_DONE, WORKER_KILL_ON_ERR, WORKER_DONE, WORKER_KILL_ON_TIMEOUT, TASK_FAILED, 0]
    )
    calc_status_list = np.repeat(calc_status_list_in, nworkers)

    # For debug
    print("Expecting: {}".format(calc_status_list))
    print("Received:  {}\n".format(H['cstat']))

    assert np.array_equal(H['cstat'], calc_status_list), "Error - unexpected calc status. Received: " + str(H['cstat'])

    # Check summary file:
    print('Checking expected task status against task summary file ...\n')

    calc_desc_list_in = [
        'Completed',
        'Worker killed task on Error',
        'Completed',
        'Worker killed task on Timeout',
        'Task Failed',
        'Manager killed on finish',
    ]

    # Repeat N times for N workers and insert Completed at start for generator
    calc_desc_list = ['Completed'] + calc_desc_list_in * nworkers
    # script_name = os.path.splitext(os.path.basename(__file__))[0]
    # short_name = script_name.split("test_", 1).pop()
    # summary_file_name = short_name + '.libe_summary.txt'
    # with open(summary_file_name,'r') as f:
    #     i=0
    #     for line in f:
    #         if "Status:" in line:
    #             _, file_status = line.partition("Status:")[::2]
    #             print("Expected: {}   Filestatus: {}".format(calc_desc_list[i], file_status.strip()))
    #             assert calc_desc_list[i] == file_status.strip(), "Status does not match file"
    #             i+=1

    print("\n\n\nRun completed.")



"""
@Author: Created by Caleb Ju at Argonne National Labs as Given's associate,
         Summer 2021.
"""
import numpy as np
import scipy.optimize as sciopt

from libensemble.message_numbers import STOP_TAG, PERSIS_STOP, FINISHED_PERSISTENT_GEN_TAG
from libensemble.tools.consensus_subroutines import (print_final_score, get_grad, get_func)


def independent_optimize(H, persis_info, gen_specs, libE_info):
    """ Uses scipy.optimize to solve objective function
    """
    ub = gen_specs['user']['ub']
    lb = gen_specs['user']['lb']

    eps = persis_info['params']['eps']
    f_i_idxs = persis_info['f_i_idxs']

    def _f(x):
        tag, f_val = get_func(x, f_i_idxs, gen_specs, libE_info)
        if tag in [STOP_TAG, PERSIS_STOP]:
            return np.inf
        return f_val

    def _df(x):
        tag, gradf_val = get_grad(x, f_i_idxs, gen_specs, libE_info)
        if tag in [STOP_TAG, PERSIS_STOP]:
            # Trick optimizer to think we found minimum
            return np.zeros(len(x))
        return gradf_val

    while 1:
        x0 = persis_info['rand_stream'].uniform(low=lb, high=ub)

        res = sciopt.minimize(_f, x0, jac=_df, method="BFGS", tol=eps,
                              options={'gtol': eps, 'norm': np.inf, 'maxiter': None})
        print_final_score(res.x, f_i_idxs, gen_specs, libE_info)

        start_pt, end_pt = f_i_idxs[0], f_i_idxs[-1]
        print('[Worker {}]: x={}'.format(persis_info['worker_num'], res.x[2*start_pt:2*end_pt]), flush=True)
        """
        try:
           res = sciopt.minimize(_f, x0, jac=_df, method="BFGS", tol=eps,
                                  options={'gtol': eps, 'norm': np.inf, 'maxiter': None})
           print_final_score(res.x, f_i_idxs, gen_specs, libE_info)

        except:
           print('hi', flush=True)
           return None, persis_info, FINISHED_PERSISTENT_GEN_TAG
        """

        return None, persis_info, FINISHED_PERSISTENT_GEN_TAG



"""
Tests libEnsemble with a uniform sample that is also requesting cancellation of
some points.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_uniform_sampling_cancel.py
   python3 test_uniform_sampling_cancel.py --nworkers 3 --comms local
   python3 test_uniform_sampling_cancel.py --nworkers 3 --comms tcp

The number of concurrent evaluations of the objective function will be 4-1=3.

Tests sampling with cancellations.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local
# TESTSUITE_NPROCS: 2 4

import numpy as np
import gc

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.six_hump_camel import six_hump_camel
from libensemble.gen_funcs.sampling import uniform_random_sample_cancel
from libensemble.tools import parse_args, add_unique_random_streams
from libensemble.tests.regression_tests.support import six_hump_camel_minima as minima

from libensemble.alloc_funcs.give_sim_work_first import give_sim_work_first as gswf
from libensemble.alloc_funcs.fast_alloc import give_sim_work_first as fast_gswf
from libensemble.alloc_funcs.only_one_gen_alloc import ensure_one_active_gen
from libensemble.alloc_funcs.give_pregenerated_work import give_pregenerated_sim_work


def create_H0(persis_info, gen_specs, sim_max):
    """Create an H0 for give_pregenerated_sim_work"""

    # Manually creating H0
    ub = gen_specs['user']['ub']
    lb = gen_specs['user']['lb']
    n = len(lb)
    b = sim_max

    H0 = np.zeros(b, dtype=[('x', float, 2), ('sim_id', int), ('given', bool), ('cancel_requested', bool)])
    H0['x'] = persis_info[0]['rand_stream'].uniform(lb, ub, (b, n))
    H0['sim_id'] = range(b)
    H0['given'] = False
    for i in range(b):
        if i % 10 == 0:
            H0[i]['cancel_requested'] = True

    # Using uniform_random_sample_cancel call - need to adjust some gen_specs though
    # gen_specs['out'].append(('sim_id', int))
    # gen_specs['out'].append(('given', bool))
    # gen_specs['user']['gen_batch_size'] = sim_max
    # H0, persis_info[0] = uniform_random_sample_cancel({}, persis_info[0], gen_specs, {})
    # H0['sim_id'] = range(gen_specs['user']['gen_batch_size'])
    # H0['given'] = False
    return H0


nworkers, is_manager, libE_specs, _ = parse_args()

sim_specs = {
    'sim_f': six_hump_camel,  # Function whose output is being minimized
    'in': ['x'],  # Keys to be given to sim_f
    'out': [('f', float)],  # Name of the outputs from sim_f
}
# end_sim_specs_rst_tag

# Note that it is unusual to specify cancel_requested as gen_specs['out']. It is here
# so that cancellations are combined with regular generator outputs for testing purposes.
# For a typical use case see test_persistent_surmise_calib.py.
gen_specs = {
    'gen_f': uniform_random_sample_cancel,  # Function generating sim_f input
    'out': [('x', float, (2,)), ('cancel_requested', bool)],
    'user': {
        'gen_batch_size': 50,  # Used by this specific gen_f
        'lb': np.array([-3, -2]),  # Used by this specific gen_f
        'ub': np.array([3, 2]),  # Used by this specific gen_f
    },
}
# end_gen_specs_rst_tag

persis_info = add_unique_random_streams({}, nworkers + 1)
sim_max = 500
exit_criteria = {'sim_max': sim_max, 'elapsed_wallclock_time': 300}

aspec1 = {
    'alloc_f': gswf,
    'out': [],
    'user': {
        'batch_mode': True,
        'num_active_gens': 1,
    },
}

aspec2 = {
    'alloc_f': gswf,
    'out': [],
    'user': {
        'batch_mode': True,
        'num_active_gens': 2,
    },
}

aspec3 = {
    'alloc_f': fast_gswf,
    'out': [],
    'user': {},
}

aspec4 = {
    'alloc_f': ensure_one_active_gen,
    'out': [],
    'user': {},
}

aspec5 = {
    'alloc_f': give_pregenerated_sim_work,
    'out': [],
    'user': {},
}

allocs = {1: aspec1, 2: aspec2, 3: aspec3, 4: aspec4, 5: aspec5}

if is_manager:
    print('Testing cancellations with non-persistent gen functions')

for testnum in range(1, 6):
    alloc_specs = allocs[testnum]
    if is_manager:
        print('\nRunning with alloc specs', alloc_specs, flush=True)

    if alloc_specs['alloc_f'] == give_pregenerated_sim_work:
        H0 = create_H0(persis_info, gen_specs, sim_max)
    else:
        H0 = None

    # Reset for those that use them
    persis_info['next_to_give'] = 0
    persis_info['total_gen_calls'] = 0  # 1

    # Perform the run - do not overwrite persis_info
    H, persis_out, flag = libE(
        sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs=libE_specs, H0=H0
    )

    if is_manager:
        assert flag == 0
        assert np.all(H['cancel_requested'][::10]), 'Some values should be cancelled but are not'
        assert np.all(~H['given'][::10]), 'Some values are given that should not have been'
        tol = 0.1
        for m in minima:
            assert np.min(np.sum((H['x'] - m) ** 2, 1)) < tol

        print("libEnsemble found the 6 minima within a tolerance " + str(tol))
        del H
        gc.collect()  # Clean up memory space.



"""
This module launches and controls the running of tasks with Balsam.

.. note:: Balsam is supported only when using ``mpi`` comms and requires Python 3.6 or higher.

In order to create a Balsam executor, the calling script should contain ::

    exctr = BalsamMPIExecutor()

The Balsam executor inherits from the MPI executor. See the
:doc:`MPIExecutor<mpi_executor>` for shared API. Any differences are
shown below.

"""

import os
import logging
import time
import datetime

from libensemble.resources import mpi_resources
from libensemble.executors.executor import \
    Application, Task, ExecutorException, TimeoutExpired, jassert, STATES
from libensemble.executors.mpi_executor import MPIExecutor

import balsam.launcher.dag as dag
from balsam.core import models

logger = logging.getLogger(__name__)
# To change logging level for just this module
# logger.setLevel(logging.DEBUG)


class BalsamTask(Task):
    """Wraps a Balsam Task from the Balsam service

    The same attributes and query routines are implemented.

    """

    def __init__(self, app=None, app_args=None, workdir=None,
                 stdout=None, stderr=None, workerid=None):
        """Instantiate a new BalsamTask instance.

        A new BalsamTask object is created with an id, status and
        configuration attributes.  This will normally be created by the
        executor on a submission.
        """
        # May want to override workdir with Balsam value when it exists
        Task.__init__(self, app, app_args, workdir, stdout, stderr, workerid)

    def read_file_in_workdir(self, filename):
        return self.process.read_file_in_workdir(filename)

    def read_stdout(self):
        return self.process.read_file_in_workdir(self.stdout)

    def read_stderr(self):
        return self.process.read_file_in_workdir(self.stderr)

    def _get_time_since_balsam_submit(self):
        """Return time since balsam task entered RUNNING state"""

        # If wait_on_start then can could calculate runtime same a base executor
        # but otherwise that will return time from task submission. Get from Balsam.

        # self.runtime = self.process.runtime_seconds # Only reports at end of run currently
        balsam_launch_datetime = self.process.get_state_times().get('RUNNING', None)
        current_datetime = datetime.datetime.now()
        if balsam_launch_datetime:
            return (current_datetime - balsam_launch_datetime).total_seconds()
        else:
            return 0

    def calc_task_timing(self):
        """Calculate timing information for this task"""

        # Get runtime from Balsam
        self.runtime = self._get_time_since_balsam_submit()

        if self.submit_time is None:
            logger.warning("Cannot calc task total_time - submit time not set")
            return

        if self.total_time is None:
            self.total_time = time.time() - self.submit_time

    def _set_complete(self, dry_run=False):
        """Set task as complete"""
        self.finished = True
        if dry_run:
            self.success = True
            self.state = 'FINISHED'
        else:
            balsam_state = self.process.state
            self.workdir = self.workdir or self.process.working_directory
            self.calc_task_timing()
            self.success = (balsam_state == 'JOB_FINISHED')
            if balsam_state == 'JOB_FINISHED':
                self.state = 'FINISHED'
            elif balsam_state == 'PARENT_KILLED':  # Not currently used
                self.state = 'USER_KILLED'
            elif balsam_state in STATES:  # In my states
                self.state = balsam_state
            else:
                logger.warning("Task finished, but in unrecognized "
                               "Balsam state {}".format(balsam_state))
                self.state = 'UNKNOWN'

            logger.info("Task {} ended with state {}".
                        format(self.name, self.state))

    def poll(self):
        """Polls and updates the status attributes of the supplied task"""
        if self.dry_run:
            return

        if not self._check_poll():
            return

        # Get current state of tasks from Balsam database
        self.process.refresh_from_db()
        balsam_state = self.process.state
        self.runtime = self._get_time_since_balsam_submit()

        if balsam_state in models.END_STATES:
            self._set_complete()

        elif balsam_state in models.ACTIVE_STATES:
            self.state = 'RUNNING'
            self.workdir = self.workdir or self.process.working_directory

        elif (balsam_state in models.PROCESSABLE_STATES or
              balsam_state in models.RUNNABLE_STATES):
            self.state = 'WAITING'

        else:
            raise ExecutorException(
                "Task state returned from Balsam is not in known list of "
                "Balsam states. Task state is {}".format(balsam_state))

    def wait(self, timeout=None):
        """Waits on completion of the task or raises TimeoutExpired exception

        Status attributes of task are updated on completion.

        Parameters
        ----------

        timeout:
            Time in seconds after which a TimeoutExpired exception is raised"""

        if self.dry_run:
            return

        if not self._check_poll():
            return

        # Wait on the task
        start = time.time()
        self.process.refresh_from_db()
        while self.process.state not in models.END_STATES:
            time.sleep(0.2)
            self.process.refresh_from_db()
            if timeout and time.time() - start > timeout:
                self.runtime = self._get_time_since_balsam_submit()
                raise TimeoutExpired(self.name, timeout)

        self.runtime = self._get_time_since_balsam_submit()
        self._set_complete()

    def kill(self, wait_time=None):
        """ Kills or cancels the supplied task """

        dag.kill(self.process)

        # Could have Wait here and check with Balsam its killed -
        # but not implemented yet.

        logger.info("Killing task {}".format(self.name))
        self.state = 'USER_KILLED'
        self.finished = True
        self.calc_task_timing()


class BalsamMPIExecutor(MPIExecutor):
    """Inherits from MPIExecutor and wraps the Balsam task management service

    .. note::  Task kills are not configurable in the Balsam executor.

    """
    def __init__(self, custom_info={}):
        """Instantiate a new BalsamMPIExecutor instance.

        A new BalsamMPIExecutor object is created with an application
        registry and configuration attributes
        """

        if custom_info:
            logger.warning("The Balsam executor does not support custom_info - ignoring")

        super().__init__(custom_info)

        self.workflow_name = "libe_workflow"

    def serial_setup(self):
        """Balsam serial setup includes empyting database and adding applications"""
        BalsamMPIExecutor.del_apps()
        BalsamMPIExecutor.del_tasks()

        for app in self.apps.values():
            calc_name = app.gname
            desc = app.desc
            full_path = app.full_path
            self.add_app(calc_name, full_path, desc)

    @staticmethod
    def del_apps():
        """Deletes all Balsam apps in the libe_app namespace"""
        AppDef = models.ApplicationDefinition

        # Some error handling on deletes.... is it internal
        for app_type in [Application.prefix]:
            deletion_objs = AppDef.objects.filter(name__contains=app_type)
            if deletion_objs:
                for del_app in deletion_objs.iterator():
                    logger.debug("Deleting app {}".format(del_app.name))
                deletion_objs.delete()

    @staticmethod
    def del_tasks():
        """Deletes all Balsam tasks """
        for app_type in [Task.prefix]:
            deletion_objs = models.BalsamJob.objects.filter(
                name__contains=app_type)
            if deletion_objs:
                for del_task in deletion_objs.iterator():
                    logger.debug("Deleting task {}".format(del_task.name))
                deletion_objs.delete()

    @staticmethod
    def add_app(name, exepath, desc):
        """ Add application to Balsam database """
        AppDef = models.ApplicationDefinition
        app = AppDef()
        app.name = name
        app.executable = exepath
        app.description = desc
        # app.default_preprocess = '' # optional
        # app.default_postprocess = '' # optional
        app.save()
        logger.debug("Added App {}".format(app.name))

    def set_resources(self, resources):
        self.resources = resources

    def submit(self, calc_type=None, app_name=None, num_procs=None,
               num_nodes=None, procs_per_node=None, machinefile=None,
               app_args=None, stdout=None, stderr=None, stage_inout=None,
               hyperthreads=False, dry_run=False, wait_on_start=False,
               extra_args=''):
        """Creates a new task, and either executes or schedules to execute
        in the executor

        The created task object is returned.
        """

        if app_name is not None:
            app = self.get_app(app_name)
        elif calc_type is not None:
            app = self.default_app(calc_type)
        else:
            raise ExecutorException("Either app_name or calc_type must be set")

        # Specific to this class
        if machinefile is not None:
            logger.warning("machinefile arg ignored - not supported in Balsam")
            jassert(num_procs or num_nodes or procs_per_node,
                    "No procs/nodes provided - aborting")

        num_procs, num_nodes, procs_per_node = \
            mpi_resources.task_partition(num_procs, num_nodes, procs_per_node)

        if stdout is not None or stderr is not None:
            logger.warning("Balsam does not currently accept a stdout "
                           "or stderr name - ignoring")
            stdout = None
            stderr = None

        # Will be possible to override with arg when implemented
        # (or can have option to let Balsam assign)
        default_workdir = os.getcwd()
        task = BalsamTask(app, app_args, default_workdir,
                          stdout, stderr, self.workerID)

        add_task_args = {'name': task.name,
                         'workflow': self.workflow_name,
                         'user_workdir': default_workdir,
                         'application': app.gname,
                         'args': task.app_args,
                         'num_nodes': num_nodes,
                         'procs_per_node': procs_per_node,
                         'mpi_flags': extra_args}

        if stage_inout is not None:
            # For now hardcode staging - for testing
            add_task_args['stage_in_url'] = "local:" + stage_inout + "/*"
            add_task_args['stage_out_url'] = "local:" + stage_inout
            add_task_args['stage_out_files'] = "*.out"

        if dry_run:
            task.dry_run = True
            logger.info('Test (No submit) Runline: {}'.format(' '.join(add_task_args)))
            task._set_complete(dry_run=True)
        else:
            task.process = dag.add_job(**add_task_args)

            if (wait_on_start):
                self._wait_on_start(task)

            if not task.timer.timing:
                task.timer.start()
                task.submit_time = task.timer.tstart  # Time not date - may not need if using timer.

            logger.info("Added task to Balsam database {}: "
                        "nodes {} ppn {}".
                        format(task.name, num_nodes, procs_per_node))

            # task.workdir = task.process.working_directory  # Might not be set yet!
        self.list_of_tasks.append(task)
        return task



#!/usr/bin/env python
import numpy as np
import sys
import matplotlib.pyplot as plt

# The following is not explicitly called but is needed for 3d plotting to work
# with older versions of python/matplotlib. It is not needed for python3.8 with
# matplotlib version 3.2.1.
from mpl_toolkits import mplot3d  # noqa


# Loop through objective points in f and extract the Pareto front.
# input: f is a list (dimensions n X p) of n p-dimensional objective points.
# output: a new list (dimensions m X p, m<n) of m nondominate points from f.
def get_pf(f):
    eps = 10.0 ** (-8.0)
    [n, p] = np.shape(f_vals)
    pareto_f = []
    # Loop over all entries in f
    for p in f_vals:
        dom = False
        # Check whether any q in f_vals dominates p
        for q in f_vals:
            # If they are equal, there is no dominance
            if (np.linalg.norm(p-q) < eps):
                continue
            # If q dominates p + a small perturbation, there is dominance
            elif ((q < p + eps).all()):
                dom = True
        # Add nondominated points to the list
        if (not dom):
            pareto_f.append(p)
    # Return the nondominated set
    return np.asarray(pareto_f)


# Load the data
if len(sys.argv) > 1:
    hist = np.load(sys.argv[1])
else:
    print('You need to supply an .npy file - aborting')
    sys.exit()

# Get first n results
if len(sys.argv) > 2:
    n = int(sys.argv[2])
    f_vals = hist['f'][:n]
    pts = get_pf(f_vals)
else:
    print('You need to supply the budget - aborting')
    sys.exit()

# Plot the contents of argv[1] when P=3.
obj1 = []
obj2 = []
obj3 = []
for i in range(pts[:, 0].size):
    obj1.append(pts[i, 0])
    obj2.append(pts[i, 1])
    obj3.append(pts[i, 2])

# Set up the figure environment.
fig = plt.figure()
ax = plt.axes(projection="3d")

# Scatter the Pareto points.
ax.scatter(obj1, obj2, obj3)
ax.set_xlabel('F1')
ax.set_ylabel('F2')
ax.set_zlabel('F3')

# Print the title.
plt.title('Tradeoff curve between objectives F1, F2, and F3')

# Display the figure.
plt.show()



#!/usr/bin/env python

'''Script to compare libEnsemble history arrays in files.

E.g., ./compare_npy.py out1.npy out2.npy

If two *.npy files are provided they are compared with each other.
If one *.npy file is provided if is compared with a hard-coded expected file
(by default located at ../expected.npy)

Default tolerances used for comparison are (rtol=1e-05, atol=1e-08). These
can be overwritten with -r (--rtol) and -a (--atol) flags.

E.g., ./compare_npy.py out1.npy out2.npy -r 1e-03

Nans compare as equal. Variable fields (such as those containing a time)
are ignored. In some cases you may have to ignore further user-defined fields

'''
import sys
import numpy as np
import argparse

desc = "Script to compare libEnsemble history arrays in files"
example = '''examples:

 ./compare_npy.py out1.npy out2.npy
 ./compare_npy.py out1.npy out2.npy --rtol 1e-03 --atol 1e-06
 '''

exclude_fields = ['gen_worker', 'sim_worker', 'gen_time', 'given_time']  # list of fields to ignore
locate_mismatch = True

parser = argparse.ArgumentParser(description=desc, epilog=example,
                                 formatter_class=argparse.RawDescriptionHelpFormatter)
parser.add_argument('-r', '--rtol', dest='rtol', type=float, default=1e-05, help='rel. tolerance')
parser.add_argument('-a', '--atol', dest='atol', type=float, default=1e-08, help='abs. tolerance')
parser.add_argument('args', nargs='*', help='*.npy files to compare')
args = parser.parse_args()

rtol = args.rtol
atol = args.atol
files = args.args

if len(files) >= 1:
    results = np.load(files[0])
    exp_results = np.load(files[1]) if len(files) >= 2 else np.load("../expected.npy")
else:
    parser.print_help()
    sys.exit()

compare_fields = tuple(filter(lambda x: x not in exclude_fields, exp_results.dtype.names))
match = all([np.allclose(exp_results[name], results[name],
                         rtol=rtol, atol=atol, equal_nan=True)
            for name in compare_fields])

print('Compare results: {}\n'.format(match))

if not locate_mismatch:
    assert match, 'Error: Results do NOT match'

if not match:
    for name in compare_fields:
        for i in range(len(results)):
            assert np.allclose(exp_results[name][i],
                   results[name][i], rtol=rtol, atol=atol, equal_nan=True), \
                'Mismatch in row ' + str(i) + ' field: ' + name + '. ' \
                + str(exp_results[name][i]) + ' ' + str(results[name][i])



"""
libEnsemble communication interface
====================================================

A comm provides a message passing abstraction for communication
between a worker user function and the manager.  Basic messages are:

  stop() - manager tells persistent gen/sim to stop
  worker_avail(nworker) - manager tells gen that workers are available
  request(recs) - worker requests simulations
  queued(id) - manager assigns simulation IDs to request
  kill(id) - gen requests manager kill a simulation
  update(id, rec) - manager informs gen of partial sim information
  result(id, rec) - manager informs gen a sim completed
  killed(id) - manager informs gen a sim was killed

To facilitate information sharing, we also have messages for history
access and monitoring (for persistent gens):

  get_history(lo, hi) - gen requests history
  subscribe() - gen subscribes to all history updates

"""

from abc import ABC, abstractmethod
from time import time
from threading import Thread
from multiprocessing import Process, Queue, Value, Lock
from traceback import format_exc
import queue
import copy

import numpy as np


class Timeout(Exception):
    """Communication timeout exception."""


class CommFinishedException(Exception):
    """Read from terminated comm exception."""


class ManagerStop(Exception):
    """Exception raised by default when manager sends a stop message."""


class RemoteException(Exception):
    """Exception raised when we received a remote exception."""

    def __init__(self, msg, exc):
        super().__init__(msg)
        self.exc = exc


class CommResult:
    """Container for a result returned on exit."""

    def __init__(self, value):
        self.value = value


class CommResultErr:
    """Container for an exception returned on exit."""

    def __init__(self, msg, exc):
        self.msg = msg
        self.exc = exc


def _timeout_fun(timeout):
    """Return a function that gets timeouts for time remaining."""
    if timeout is None:
        return (lambda: None)
    else:
        tdeath = time() + timeout
        return (lambda: tdeath-time())


class Comm(ABC):
    """Bidirectional communication
    """

    @abstractmethod
    def send(self, *args):
        """Send a message."""

    @abstractmethod
    def recv(self, timeout=None):
        """Receive a message or raise TimeoutError."""

    def mail_flag(self):
        """Check whether we know a message is ready for receipt."""
        return False

    def kill_pending(self):
        """Cancel any pending sends (don't worry about those in the system)."""


class QComm(Comm):
    """Queue-based bidirectional communicator

    A QComm provides a message passing abstraction based on a pair of message
    queues: an inbox for incoming messages and an outbox for outgoing messages.
    These can be used with threads or multiprocessing.
    """

    # Integer count - shared amongst processes
    lock = Lock()
    _ncomms = Value('i', 0)

    def __init__(self, inbox, outbox, copy_msg=False):
        """Set the inbox and outbox queues."""
        self._inbox = inbox
        self._outbox = outbox
        self._copy = copy_msg
        self.recv_buffer = None

    def get_num_workers(self):
        """Return global _ncomms"""
        return QComm._ncomms.value

    def send(self, *args):
        """Place a message on the outbox queue."""
        if self._copy:
            args = copy.deepcopy(args)
        self._outbox.put(args)

    def recv(self, timeout=None):
        """Return a message from the inbox queue or raise TimeoutError."""
        pb_result = self.recv_buffer
        self.recv_buffer = None
        if pb_result is not None:
            return pb_result
        try:
            if not self._inbox.empty():
                return self._inbox.get()
            return self._inbox.get(timeout=timeout)
        except queue.Empty:
            raise Timeout()

    # TODO: This should go away once I have internal comms working
    def push_to_buffer(self, *args):
        self.recv_buffer = args

    def mail_flag(self):
        """Check whether we know a message is ready for receipt."""
        return not self._inbox.empty()


class QCommThread(Comm):
    """Launch a user function in a thread with an attached QComm.
    """

    def __init__(self, main, *args, **kwargs):
        self.inbox = queue.Queue()
        self.outbox = queue.Queue()
        self.main = main
        self._result = None
        self._exception = None
        kwargs['comm'] = QComm(self.inbox, self.outbox, True)
        self.thread = Thread(target=self._qcomm_main,
                             args=args, kwargs=kwargs)

    def send(self, *args):
        """Send a message to the thread (called from creator)"""
        self.inbox.put(copy.deepcopy(args))

    def recv(self, timeout=None):
        """Return a message from the thread or raise TimeoutError."""
        try:
            if not self.outbox.empty():
                return self.outbox.get()
            return self.outbox.get(timeout=timeout)
        except queue.Empty:
            raise Timeout()

    def mail_flag(self):
        """Check whether we know a message is ready for receipt."""
        return not self.outbox.empty()

    def run(self):
        """Start the thread."""
        self.thread.start()

    def result(self):
        """Join and return the thread main result (or re-raise an exception)."""
        self.thread.join()
        if isinstance(self._exception, Exception):
            raise self._exception
        return self._result

    @property
    def running(self):
        """Check if the thread is running."""
        return self.thread.is_alive()

    def _qcomm_main(self, *args, **kwargs):
        """Main routine -- handles return values and exceptions."""
        try:
            self._result = self.main(*args, **kwargs)
        except Exception as e:
            self._exception = e

    def __enter__(self):
        self.run()
        return self

    def __exit__(self, etype, value, traceback):
        self.thread.join()


class QCommProcess(Comm):
    """Launch a user function in a process with an attached QComm.
    """

    def __init__(self, main, *args, **kwargs):
        self.inbox = Queue()
        self.outbox = Queue()
        self._result = None
        self._exception = None
        self._done = False
        comm = QComm(self.inbox, self.outbox)
        with QComm.lock:
            QComm._ncomms.value += 1
        self.process = Process(target=QCommProcess._qcomm_main,
                               args=(comm, main) + args, kwargs=kwargs)

    def _is_result_msg(self, msg):
        """Return true if message indicates final result (and set result/except)."""
        if len(msg) and isinstance(msg[0], CommResult):
            self._result = msg[0].value
            self._done = True
            return True
        if len(msg) and isinstance(msg[0], CommResultErr):
            self._exception = msg[0]
            self._done = True
            return True
        return False

    def send(self, *args):
        """Send a message to the thread (called from creator)"""
        self.inbox.put(args)

    def recv(self, timeout=None):
        """Return a message from the thread or raise TimeoutError."""
        try:
            if self._done:
                raise CommFinishedException()
            if not self.outbox.empty():
                msg = self.outbox.get()
            else:
                msg = self.outbox.get(timeout=timeout)
            if self._is_result_msg(msg):
                raise CommFinishedException()
            return msg
        except queue.Empty:
            raise Timeout()

    def mail_flag(self):
        """Check whether we know a message is ready for receipt."""
        return not self.outbox.empty()

    def run(self):
        """Start the process."""
        self.process.start()

    def result(self, timeout=None):
        """Join and return the thread main result (or re-raise an exception)."""
        get_timeout = _timeout_fun(timeout)
        while not self._done and (timeout is None or timeout >= 0):
            try:
                msg = self.outbox.get(timeout=timeout)
            except queue.Empty:
                raise Timeout()
            self._is_result_msg(msg)
            timeout = get_timeout()
        if not self._done:
            raise Timeout()
        self.process.join(timeout=timeout)
        if self.running:
            raise Timeout()
        if self._exception is not None:
            raise RemoteException(self._exception.msg, self._exception.exc)
        with QComm.lock:
            QComm._ncomms.value -= 1
        return self._result

    def terminate(self, timeout=None):
        """Terminate the process."""
        if self.running:
            self.process.terminate()
        self.process.join(timeout=timeout)
        if self.running:
            raise Timeout()
        with QComm.lock:
            QComm._ncomms.value -= 1

    @property
    def running(self):
        """Return true if process is running"""
        return self.process.is_alive()

    @staticmethod
    def _qcomm_main(comm, main, *args, **kwargs):
        """Main routine -- handles return values and exceptions."""
        try:
            _result = main(comm, *args, **kwargs)
            comm.send(CommResult(_result))
        except Exception as e:
            comm.send(CommResultErr(str(e), format_exc()))
            raise e

    def __enter__(self):
        self.run()
        return self

    def __exit__(self, etype, value, traceback):
        self.process.join()


class CommHandler(ABC):
    """Comm wrapper with message handler dispatching.

    The comm wrapper defines a message processor that dispatches to
    different handler methods based on message types.  An incoming message
    with the tag 'foo' gets dispatched to a handler 'on_foo'; if 'on_foo'
    is not defined, we pass to the 'on_unhandled_message' routine.
    """

    def __init__(self, comm):
        """Set the comm to be wrapped."""
        self.comm = comm

    def send(self, *args):
        """Send via the comm."""
        self.comm.send(*args)

    def process_message(self, timeout=None):
        """Receive and process a message via the comm."""
        msg = self.comm.recv(timeout)
        msg_type = msg[0]
        args = msg[1:]
        try:
            method = 'on_{}'.format(msg_type)
            handler = getattr(self, method)
        except AttributeError:
            return self.on_unhandled_message(msg)
        return handler(*args)

    def on_unhandled_message(self, msg):
        """Handle any messages for which there are no named handlers."""
        raise ValueError("No handler available for message {0}{1}".
                         format(msg[0], msg[1:]))


class GenCommHandler(CommHandler):
    """Wrapper for handling messages at a persistent gen."""

    def send_request(self, recs):
        """Request new evaluations."""
        self.send('request', recs)

    def send_kill(self, sim_id):
        """Kill an evaluation."""
        self.send('kill', sim_id)

    def send_get_history(self, lo, hi):
        """Request history from manager."""
        self.send('get_history', lo, hi)

    def send_subscribe(self):
        """Request subscription to updates on sims not launched by this gen."""
        self.send('subscribe')

    def on_stop(self):
        """Handle stop message."""
        raise ManagerStop()

    @abstractmethod
    def on_worker_avail(self, nworker):
        """Handle updated number of workers available to perform sims."""

    @abstractmethod
    def on_queued(self, sim_id):
        """Handle sim_id assignment in response to a request"""

    @abstractmethod
    def on_result(self, sim_id, recs):
        """Handle simulation results"""

    @abstractmethod
    def on_update(self, sim_id, recs):
        """Handle simulation updates"""

    @abstractmethod
    def on_killed(self, sim_id):
        """Handle a simulation kill"""


class SimCommHandler(CommHandler):
    """Wrapper for handling messages at sim."""

    def send_result(self, sim_id, recs):
        """Send a simulation result"""
        self.send('result', sim_id, recs)

    def send_update(self, sim_id, recs):
        """Send a simulation update"""
        self.send('update', sim_id, recs)

    def send_killed(self, sim_id):
        """Send notification that a simulation was killed"""
        self.send('killed', sim_id)

    def on_stop(self):
        """Handle stop message."""
        raise ManagerStop()

    @abstractmethod
    def on_request(self, sim_id, recs):
        """Handle a request for a simulation"""

    @abstractmethod
    def on_kill(self, sim_id):
        """Handle a request to kill a simulation"""


class CommEval(GenCommHandler):
    """Future-based interface for generator comms
    """

    def __init__(self, comm, workers=0, gen_specs=None):
        super().__init__(comm)
        self.sim_started = 0
        self.sim_pending = 0
        self.workers = workers
        self.gen_specs = gen_specs
        self.promises = {}
        self.returning_promises = None
        self.waiting_for_queued = 0

    def request(self, recs):
        """Request simulations, return promises"""
        self.sim_started += len(recs)
        self.sim_pending += len(recs)
        self.send_request(recs)
        self.waiting_for_queued = len(recs)
        while self.waiting_for_queued > 0:
            self.process_message()
        returning_promises = self.returning_promises
        self.returning_promises = None
        return returning_promises

    def __call__(self, *args, **kwargs):
        """Request a simulation and return a promise"""
        assert not (args and kwargs), \
            "Must specify simulation args by position or keyword, but not both"
        assert args or kwargs, \
            "Must specify simulation arguments."
        rec = np.zeros(1, dtype=self.gen_specs['out'])
        if args:
            assert len(args) == len(self.gen_specs['out']), \
                "Wrong number of positional arguments in sim call."
            for k, spec in enumerate(self.gen_specs['out']):
                name = spec[0]
                rec[name] = args[k]
        else:
            for name, value in kwargs.items():
                rec[name] = value
        return self.request(rec)[0]

    def wait_any(self):
        """Wait for any pending simulation to be done"""
        sim_pending = self.sim_pending
        while sim_pending == self.sim_pending:
            self.process_message()

    def wait_all(self):
        """Wait for all pending simulations to be done"""
        while self.sim_pending > 0:
            self.process_message()

    # --- Message handlers

    def on_worker_avail(self, nworker):
        """Update worker count"""
        self.workers = nworker
        return -1

    def on_queued(self, sim_id):
        """Set up futures with indicated simulation IDs"""
        lo = sim_id
        hi = sim_id + self.waiting_for_queued
        self.waiting_for_queued = 0
        self.returning_promises = []
        for s in range(lo, hi):
            promise = Future(self, s)
            self.promises[s] = promise
            self.returning_promises.append(promise)
        return -1

    def on_result(self, sim_id, recs):
        """Handle completed simulation"""
        for k, rec in enumerate(recs):
            self.sim_pending -= 1
            self.promises[sim_id+k].on_result(rec)
        return sim_id

    def on_update(self, sim_id, recs):
        """Handle updated simulation"""
        for k, rec in enumerate(recs):
            self.promises[sim_id+k].on_update(rec)
        return sim_id

    def on_killed(self, sim_id):
        """Handle killed simulation"""
        self.sim_pending -= 1
        self.promises[sim_id].on_killed()
        return sim_id


class Future:
    """Future objects for monitoring asynchronous simulation calls.

    The Future objects are not meant to be instantiated on their own;
    they are only produced by a call on a CommEval object.
    """

    def __init__(self, ceval, sim_id):
        self._ceval = ceval
        self._id = sim_id
        self._comm = ceval.comm
        self._result = None
        self._killed = False
        self._success = False

    @property
    def current_result(self):
        """Return the current (possibly incomplete) result immediately."""
        return self._result

    def cancelled(self):
        """Return True if the simulation was killed."""
        return self._killed

    def done(self):
        """Return True if the simulation completed successfully or was killed."""
        return self._success or self._killed

    def cancel(self):
        """Cancel the simulation."""
        self._ceval.send_kill(self._id)

    def result(self, timeout=None):
        """Get the result of the simulation or throw a timeout."""
        while not self.done():
            if timeout is not None and timeout < 0:
                raise Timeout()
            tstart = time()
            try:
                self._ceval.process_message(timeout)
            except Timeout:
                pass
            if timeout is not None:
                timeout -= (time() - tstart)
        return self._result

    # --- Message handlers

    def on_result(self, result):
        """Handle an incoming result."""
        self._result = result
        self._success = True

    def on_update(self, result):
        """Handle an incoming update."""
        self._result = result

    def on_killed(self):
        """Handle a kill notification."""
        self._killed = True



"""
Test the libEnsemble capability to honor a generator function's request to
stop a run.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 5 python3 test_persistent_uniform_gen_decides_stop.py
   python3 test_persistent_uniform_gen_decides_stop.py --nworkers 4 --comms local
   python3 test_persistent_uniform_gen_decides_stop.py --nworkers 4 --comms tcp

The number of concurrent evaluations of the objective function with 2 gens will be 2:
5 - 1 manager - 2 persistent gens = 2.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 5

import sys
import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.branin.branin_obj import call_branin as sim_f
from libensemble.gen_funcs.persistent_uniform_sampling import persistent_request_shutdown as gen_f
from libensemble.alloc_funcs.start_only_persistent import only_persistent_gens as alloc_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")

n = 2
ngens = 2

if ngens >= nworkers:
    sys.exit("You number of generators must be less than the number of workers -- aborting...")

sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
    'user': {'uniform_random_pause_ub': 0.5},
}

gen_specs = {
    'gen_f': gen_f,
    'persis_in': ['f', 'x', 'sim_id'],
    'out': [('x', float, (n,))],
    'user': {
        'initial_batch_size': nworkers,
        'shutdown_limit': 10,  # Iterations on a gen before it triggers a shutdown.
        'lb': np.array([-3, -2]),
        'ub': np.array([3, 2]),
    },
}

alloc_specs = {
    'alloc_f': alloc_f,
    'user': {
        'async_return': True,
        'num_active_gens': ngens,
    },
}

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {'gen_max': 50, 'elapsed_wallclock_time': 300}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

if is_manager:
    [_, counts] = np.unique(H['gen_time'], return_counts=True)
    print('Num. points in each gen iteration:', counts)
    assert counts[0] == nworkers, "The first gen_time should be common among initial_batch_size number of points"
    assert counts[1] == nworkers, "The second gen_time should be common among initial_batch_size number of points"
    assert len(np.unique(counts)) > 1, "There is no variablitiy in the gen_times but there should be for the async case"

    gen_workers = np.unique(H['gen_worker'])
    print('Generators that issued points', gen_workers)
    assert len(gen_workers) == ngens, "The number of gens used {} does not match num_active_gens {}".format(
        len(gen_workers), ngens
    )

    save_libE_output(H, persis_info, __file__, nworkers)



#!/usr/bin/env python

"""
Unit test of launcher helpers for libensemble.
"""

import sys
import libensemble.utils.launcher as launcher


def test_form_command():
    "Test the command templating."

    run_specs = {'mpirun': 'mpirun', 'nproc': 10, 'nrank': 5, 'mf': None}

    cmd = ['{mpirun}', '-n {nproc}', '-nper {nrank}', '-machinefile {mf}',
           'more arguments "ho hum"']
    args = launcher.form_command(cmd, run_specs)
    aref = ['mpirun', '-n', '10', '-nper', '5', 'more', 'arguments', '"ho hum"']

    assert args == aref, "Command templating test failed."


def xtest_submit():
    "Test simple launch."

    py_exe = sys.executable or "python"

    # Launch infinite loop, pay attention to term
    process = launcher.launch([py_exe, "launch_busy.py"])
    assert not launcher.process_is_stopped(process, 0.1), "Process stopped early."
    launcher.cancel(process, 0.5)

    # Launch infinite loop, ignore term
    process = launcher.launch([py_exe, "launch_busy.py", "1"])
    assert not launcher.process_is_stopped(process, 0.5), "Process stopped early."
    launcher.cancel(process, 0.5)

    # Launch infinite loop, pay attention to term
    process = launcher.launch([py_exe, "launch_busy.py"],
                              start_new_session=True)
    assert not launcher.process_is_stopped(process, 0.1), "Process stopped early."
    launcher.cancel(process, 0.5)

    # Launch infinite loop, ignore term
    process = launcher.launch([py_exe, "launch_busy.py", "1"],
                              start_new_session=True)
    assert not launcher.process_is_stopped(process, 0.5), "Process stopped early."
    launcher.cancel(process, 0.5)

    # Check proper handling of ProcessLookupError
    assert not launcher.killpg(process), "Expected lookup error."
    assert not launcher.terminatepg(process), "Expected lookup error."

    # Launch finite loop, wait for termination
    process = launcher.launch([py_exe, "launch_busy.py", "0", "0.1"])
    assert launcher.process_is_stopped(process, 1.5), "Process should have stopped earlier."

    # Try simple kill
    process = launcher.launch([py_exe, "launch_busy.py", "1"])
    assert not launcher.process_is_stopped(process, 0.5), "Process stopped early."
    launcher.cancel(process, 0)


def test_launch32():
    "If we are in Python > 3.2, still check that 3.2 wait func works"
    saved_wait = launcher.wait
    launcher.wait = launcher.wait_py32
    xtest_submit()
    launcher.wait = saved_wait


def test_launch33():
    "If we are in Python > 3.2, also check the new-style wait func"
    if launcher.wait == launcher.wait_py33:
        xtest_submit()



import numpy as np
from libensemble.message_numbers import EVAL_GEN_TAG
from libensemble.tools.alloc_support import AllocSupport, InsufficientFreeResources


def only_persistent_gens_for_inverse_bayes(W, H, sim_specs, gen_specs, alloc_specs, persis_info, libE_info):
    """
    Starts up to gen_count number of persistent generators.
    These persistent generators produce points (x) in batches and subbatches.
    The points x are given in subbatches to workers to perform a calculation.
    When all subbatches have returned, their output is given back to the
    corresponding persistent generator.

    The first time called there are no persis_w 1st for loop is not done
    """

    if libE_info['sim_max_given'] or not libE_info['any_idle_workers']:
        return {}, persis_info

    user = alloc_specs.get('user', {})
    sched_opts = user.get('scheduler_opts', {})
    manage_resources = 'resource_sets' in H.dtype.names or libE_info['use_resource_sets']

    support = AllocSupport(W, manage_resources, persis_info, sched_opts)
    Work = {}
    gen_count = support.count_persis_gens()

    # If wid is idle, but in persistent mode, and generated work has all returned
    # give output back to wid. Otherwise, give nothing to wid
    for wid in support.avail_worker_ids(persistent=EVAL_GEN_TAG):

        # if > 1 persistent generator, assign the correct work to it
        inds_generated_by_wid = (H['gen_worker'] == wid)
        if support.all_returned(H, inds_generated_by_wid):

            # Has sim_f completed everything from this persistent worker?
            # Then give back everything in the last batch
            batch_ids = H['batch'][inds_generated_by_wid]
            last_batch_inds = (batch_ids == np.max(batch_ids))
            inds_to_send_back = np.where(np.logical_and(inds_generated_by_wid, last_batch_inds))[0]
            if H['batch'][-1] > 0:
                n = gen_specs['user']['subbatch_size'] * gen_specs['user']['num_subbatches']
                k = H['batch'][-1]
                H['weight'][(n * (k - 1)):(n * k)] = H['weight'][(n * k):(n * (k + 1))]

            Work[wid] = support.gen_work(wid, ['like'], inds_to_send_back, persis_info.get(wid), persistent=True)

    points_to_evaluate = ~H['given'] & ~H['cancel_requested']
    for wid in support.avail_worker_ids(persistent=False):
        if np.any(points_to_evaluate):

            # perform sim evaluations (if any point hasn't been given).
            sim_subbatches = H['subbatch'][points_to_evaluate]
            sim_inds = (sim_subbatches == np.min(sim_subbatches))
            sim_ids_to_send = np.nonzero(points_to_evaluate)[0][sim_inds]

            try:
                Work[wid] = support.sim_work(wid, H, sim_specs['in'], sim_ids_to_send, [])
            except InsufficientFreeResources:
                break
            points_to_evaluate[sim_ids_to_send] = False

        elif gen_count == 0:

            # Finally, generate points since there is nothing else to do.
            try:
                Work[wid] = support.gen_work(wid, gen_specs['in'], [], persis_info.get(wid), persistent=True)
            except InsufficientFreeResources:
                break
            gen_count += 1

    return Work, persis_info



#!/usr/bin/env python
import sys
import pickle
from pprint import pprint

pretty = True  # Default

if len(sys.argv) > 1:
    filename = sys.argv[1]
else:
    print('You need to supply an .pickle file - aborting')
    sys.exit()

if len(sys.argv) > 2:
    if sys.argv[2] == 'pretty' or 'p':
        pretty = True
    if sys.argv[2] == 'raw' or 'r':
        pretty = False

with open(filename, 'rb') as f:
    data = pickle.load(f)

if pretty:
    pprint(data, compact=True)
else:
    print(data)



"""
Test of MPI comms.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_mpi_comms.py

The number of concurrent evaluations of the objective function will be 4-1=3.
"""

from mpi4py import MPI
from libensemble.comms.mpi import MPIComm, Timeout
from libensemble.tools import parse_args

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi
# TESTSUITE_NPROCS: 2 4

nworkers, is_manager, libE_specs, _ = parse_args()

assert libE_specs['comms'] == 'mpi', "This test can only be run with mpi comms -- aborting..."


def check_recv(comm, expected_msg):
    msg = comm.recv()
    assert msg == expected_msg, "Expected {}, received {}".format(expected_msg, msg)


def worker_main(mpi_comm):
    "Worker main routine"
    comm = MPIComm(mpi_comm)
    check_recv(comm, "Hello")
    check_recv(comm, "World")
    check_recv(comm, comm.rank)
    comm.send(comm.rank)
    check_recv(comm, "Goodbye")


def manager_main(mpi_comm):
    "Manager main routine"
    worker_comms = [MPIComm(mpi_comm, r) for r in range(1, mpi_comm.Get_size())]
    for comm in worker_comms:
        try:
            okay_flag = True
            comm.recv(0)
            okay_flag = False
        except Timeout:
            pass
        assert okay_flag, "Worker comm behaved as expected."
    for comm in worker_comms:
        comm.send("Hello")
        comm.send("World")
        comm.send(comm.remote_rank)
    for comm in worker_comms:
        check_recv(comm, comm.remote_rank)
        comm.send("Goodbye")


def mpi_comm_excl(exc=[0]):
    world_group = MPI.COMM_WORLD.Get_group()
    new_group = world_group.Excl(exc)
    mpi_comm = MPI.COMM_WORLD.Create(new_group)
    return mpi_comm


def check_ranks(mpi_comm, test_exp, test_num):
    try:
        rank = mpi_comm.Get_rank()
    except Exception:
        rank = -1
    comm_ranks_in_world = MPI.COMM_WORLD.allgather(rank)
    print('got {},  exp {} '.format(comm_ranks_in_world, test_exp[test_num]), flush=True)
    # This is really testing the test is testing what is it supposed to test
    assert comm_ranks_in_world == test_exp[test_num], (
        "comm_ranks_in_world are: " + str(comm_ranks_in_world) + " Expected: " + str(test_exp[test_num])
    )
    if rank == -1:
        return False
    return True


# Run Tests
all_ranks = list(range(MPI.COMM_WORLD.Get_size()))

tests = {
    1: MPI.COMM_WORLD.Dup,
    2: mpi_comm_excl,
}

test_exp = {
    1: all_ranks,
    2: [-1] + all_ranks[:-1],
}

for test_num in range(1, len(tests) + 1):
    mpi_comm = tests[test_num]()
    if check_ranks(mpi_comm, test_exp, test_num):
        is_manager = mpi_comm.Get_rank() == 0
        if is_manager:
            manager_main(mpi_comm)
        else:
            worker_main(mpi_comm)



"""Contains supplemental methods for gen function in persistent_surmise_calib.py."""

import numpy as np
import scipy.stats as sps


class thetaprior:
    """Define the class instance of priors provided to the methods."""
    def __init__(self, loc, scale):
        self._loc = loc
        self._scale = scale

    def lpdf(self, theta):
        """Return log prior density."""
        loc = self._loc
        scale = self._scale
        return (np.sum(sps.norm.logpdf(theta, loc, scale), 1)).reshape((len(theta), 1))

    def rnd(self, n):
        """Return random draws from prior."""
        loc = self._loc
        scale = self._scale
        return np.vstack((sps.norm.rvs(loc, scale, size=(n, 4))))


def gen_true_theta():
    """Generate one parameter to be the true parameter for calibration."""
    theta0 = np.atleast_2d(np.array([0.5] * 4))

    return theta0


def gen_thetas(prior, n):
    """Generate and return n parameters for the test function."""
    thetas = prior.rnd(n)
    return thetas


def gen_xs(nx, randstream):
    """Generate and returns n inputs for the modified Borehole function."""
    xs = randstream.uniform(0, 1, (nx, 3))
    xs[:, 2] = xs[:, 2] > 0.5

    return xs


def gen_observations(fevals, obsvar_const, randstream):
    """Generate observations."""
    n_x = len(np.squeeze(fevals))
    obsvar = np.maximum(obsvar_const*fevals, 5)
    obsvar = np.squeeze(obsvar)
    obs = fevals + randstream.normal(0, np.sqrt(obsvar), n_x).reshape((n_x))

    obs = np.squeeze(obs)
    return obs, obsvar


def select_next_theta(numnewtheta, cal, emu, pending, numexplore):
    # numnewtheta += 2
    thetachoices = cal.theta(numexplore)
    choicescost = np.ones(thetachoices.shape[0])
    thetaneworig, info = emu.supplement(size=numnewtheta, thetachoices=thetachoices,
                                        choicescost=choicescost,
                                        cal=cal, overwrite=True,
                                        args={'includepending': True,
                                              'costpending': 0.01 + 0.99 * np.mean(pending, 0),
                                              'pending': pending})
    thetaneworig = thetaneworig[:numnewtheta, :]
    thetanew = thetaneworig
    return thetanew, info


def obviate_pend_theta(info, pending):
    return pending, info['obviatesugg']



"""
Runs libEnsemble with Latin hypercube sampling on a simple 1D problem using
the libEnsemble yaml interface

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_1d_sampling_from_yaml.py
   python3 test_1d_sampling_from_yaml.py --nworkers 3 --comms local
   python3 test_1d_sampling_from_yaml.py --nworkers 3 --comms tcp

The number of concurrent evaluations of the objective function will be 4-1=3.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 2 4

import numpy as np
from libensemble import Ensemble

sampling = Ensemble()
sampling.from_yaml('1d_sampling.yaml')

sampling.gen_specs['user'].update(
    {
        'lb': np.array([-3]),
        'ub': np.array([3]),
    }
)

sampling.persis_info.add_random_streams()

# Perform the run
sampling.run()

if sampling.is_manager:
    assert len(sampling.H) >= 501
    print("\nlibEnsemble with random sampling has generated enough points")
    sampling.save_output(__file__)



import os
from libensemble.resources.env_resources import EnvResources


def setup_standalone_run():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = ""


def teardown_standalone_run():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = ""


def setup_function(function):
    print("setup_function    function:%s" % function.__name__)
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = ""


def teardown_function(function):
    print("teardown_function    function:%s" % function.__name__)
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = ""

# Tests ========================================================================================

# Tests for obtaining nodelist from environment variables


def test_slurm_nodelist_empty():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = ""
    exp_out = []  # empty
    nodelist = EnvResources.get_slurm_nodelist(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_slurm_nodelist_single():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "knl-0056"
    exp_out = ["knl-0056"]
    nodelist = EnvResources.get_slurm_nodelist(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_slurm_nodelist_single_nodash():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "nid00056"
    exp_out = ["nid00056"]
    nodelist = EnvResources.get_slurm_nodelist(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_slurm_nodelist_knl_seq():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "knl-[0009-0012]"
    exp_out = ['knl-0009', 'knl-0010', 'knl-0011', 'knl-0012']
    nodelist = EnvResources.get_slurm_nodelist(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_slurm_nodelist_bdw_seq():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "bdw-[0009-0012]"
    exp_out = ['bdw-0009', 'bdw-0010', 'bdw-0011', 'bdw-0012']
    nodelist = EnvResources.get_slurm_nodelist(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_slurm_nodelist_groups():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "knl-[0020-0022,0137-0139,1234]"
    exp_out = ['knl-0020', 'knl-0021', 'knl-0022', 'knl-0137', 'knl-0138', 'knl-0139', 'knl-1234']
    nodelist = EnvResources.get_slurm_nodelist(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_slurm_nodelist_groups_partitions():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "bdw-[0254,0384,0565-0568],bdwd-[0004,0009]"
    exp_out = ['bdw-0254', 'bdw-0384', 'bdw-0565', 'bdw-0566', 'bdw-0567', 'bdw-0568', 'bdwd-0004', 'bdwd-0009']
    nodelist = EnvResources.get_slurm_nodelist(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_slurm_nodelist_groups_nodash():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "nid0[0020-0022,0137-0139,1234]"
    exp_out = ['nid00020', 'nid00021', 'nid00022', 'nid00137', 'nid00138', 'nid00139', 'nid01234']
    nodelist = EnvResources.get_slurm_nodelist(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_slurm_nodelist_groups_longprefix():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "super-[000020-000022,000137-000139,001234,023456-023458]"
    exp_out = ['super-000020', 'super-000021', 'super-000022', 'super-000137', 'super-000138', 'super-000139',
               'super-001234', 'super-023456', 'super-023457', 'super-023458']
    nodelist = EnvResources.get_slurm_nodelist(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_slurm_nodelist_reverse_grp():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "knl-[0020-0022,0139-0137,1234]"
    exp_out = ['knl-0020', 'knl-0021', 'knl-0022', 'knl-0137', 'knl-0138', 'knl-0139', 'knl-1234']
    nodelist = EnvResources.get_slurm_nodelist(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_cobalt_nodelist_empty():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = ""
    exp_out = []  # empty
    nodelist = EnvResources.get_cobalt_nodelist(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_cobalt_nodelist_single():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "56"
    exp_out = ["56"]
    nodelist = EnvResources.get_cobalt_nodelist(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_cobalt_nodelist_seq():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "9-12"
    exp_out = ['9', '10', '11', '12']
    nodelist = EnvResources.get_cobalt_nodelist(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_cobalt_nodelist_groups():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "20-22,137-139,1234"
    exp_out = ['20', '21', '22', '137', '138', '139', '1234']
    nodelist = EnvResources.get_cobalt_nodelist(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_cobalt_nodelist_reverse_grp():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "20-22,139-137,1234"
    exp_out = ['20', '21', '22', '137', '138', '139', '1234']
    nodelist = EnvResources.get_cobalt_nodelist(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_lsf_nodelist_empty():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = ""
    exp_out = []  # empty
    nodelist = EnvResources.get_lsf_nodelist(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_lsf_nodelist_single():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = 'batch5' + ' g06n02' * 42
    exp_out = ["g06n02"]
    nodelist = EnvResources.get_lsf_nodelist(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_lsf_nodelist_seq():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = 'batch5' + ' g06n02' * 42 + ' h21n18' * 42
    exp_out = ['g06n02', 'h21n18']
    nodelist = EnvResources.get_lsf_nodelist(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"

# These dont apply to the lsf lists as they are listed in full
# def test_lsf_nodelist_groups():
# def test_lsf_nodelist_reverse_grp():


def test_lsf_nodelist_shortform_empty():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = ""
    exp_out = []  # empty
    nodelist = EnvResources.get_lsf_nodelist_frm_shortform(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_lsf_nodelist_shortform_single():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = 'batch5 1 g06n02 42'
    exp_out = ["g06n02"]
    nodelist = EnvResources.get_lsf_nodelist_frm_shortform(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_lsf_nodelist_shortform_seq():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = 'batch5 1 g06n02 42 h21n18 42'
    exp_out = ['g06n02', 'h21n18']
    nodelist = EnvResources.get_lsf_nodelist_frm_shortform(node_list_env="LIBE_RESOURCES_TEST_NODE_LIST")
    assert nodelist == exp_out, "Nodelist returned does not match expected"


def test_shortnames_nochange_slurm():
    env_resources = EnvResources()
    # Test Slurm abbrev
    exp_names = ['knl-0019', 'knl-0021', 'knl-0022', 'knl-0137', 'knl-0138', 'knl-0139', 'knl-2345']
    env_resources.scheduler = 'Slurm'
    abbrev_names = env_resources.shortnames(exp_names)
    assert abbrev_names == exp_names, "Abbreviated names returned do not match expected"
    del env_resources


def test_shortnames_slurm():
    env_resources = EnvResources()
    # Test Slurm abbrev
    exp_names = ['knl-0019', 'knl-0021', 'knl-0022']
    full_names = ['knl-0019.some.suffix', 'knl-0021.some.suffix', 'knl-0022.diff_suffix']
    env_resources.scheduler = 'Slurm'
    abbrev_names = env_resources.shortnames(full_names)
    assert abbrev_names == exp_names, "Abbreviated names returned do not match expected"
    del env_resources


def test_shortnames_nochange_cobalt():
    env_resources = EnvResources()
    # Test Cobalt abbrev
    exp_names = ['21', '22', '137', '138', '1234', '11234']
    env_resources.scheduler = 'Cobalt'
    abbrev_names = env_resources.shortnames(exp_names)
    assert abbrev_names == exp_names, "Abbreviated names returned do not match expected"
    del env_resources


def test_shortnames_cobalt():
    env_resources = EnvResources()
    # Test Cobalt abbrev
    exp_names = ['20', '21', '22', '137', '138', '1234', '11234']
    full_names = ['nid00020', 'nid00021', 'nid00022', 'nid00137', 'nid00138', 'nid01234', 'nid11234']
    env_resources.scheduler = 'Cobalt'
    abbrev_names = env_resources.shortnames(full_names)
    assert abbrev_names == exp_names, "Abbreviated names returned do not match expected"
    del env_resources


if __name__ == "__main__":
    setup_standalone_run()

    test_slurm_nodelist_empty()
    test_slurm_nodelist_single()
    test_slurm_nodelist_single_nodash()
    test_slurm_nodelist_knl_seq()
    test_slurm_nodelist_bdw_seq()
    test_slurm_nodelist_groups()
    test_slurm_nodelist_groups_partitions()
    test_slurm_nodelist_groups_nodash()
    test_slurm_nodelist_groups_longprefix()
    test_slurm_nodelist_reverse_grp()

    test_cobalt_nodelist_empty()
    test_cobalt_nodelist_single()
    test_cobalt_nodelist_seq()
    test_cobalt_nodelist_groups()
    test_cobalt_nodelist_reverse_grp()

    test_lsf_nodelist_empty()
    test_lsf_nodelist_single()
    test_lsf_nodelist_seq()

    test_lsf_nodelist_shortform_empty()
    test_lsf_nodelist_shortform_single()
    test_lsf_nodelist_shortform_seq()

    test_shortnames_nochange_slurm()
    test_shortnames_slurm()
    test_shortnames_nochange_cobalt()
    test_shortnames_cobalt()

    teardown_standalone_run()
